"""Module for simulating isotropic tropospheric turbulence and
estimating turbulence properties.

Author: Scott Staniewicz
Adapted from MintPy fractal.py module,
  https://github.com/insarlab/MintPy/blob/4c2b6f8f80c86f245f8635bbb0ce2e46b29cb576/mintpy/simulation/fractal.py
  (Author: Zhang Yunjun, 2019)
which was a Python translation in MintPy of the matlab scripts written by
Ramon Hanssen, May 2000, available in the following website:
    http://doris.tudelft.nl/software/insarfractal.tar.gz
"""

import copy
from concurrent.futures import ProcessPoolExecutor, as_completed
from pathlib import Path
from typing import List

import numpy as np
from numpy.polynomial.polynomial import Polynomial, polyval
from numpy.random import SeedSequence, default_rng
from scipy import ndimage
from scipy.fft import fft2, fftfreq, fftshift, ifft2
from tqdm.auto import tqdm

from . import utils

MAX_WORKERS = 8
RNG = default_rng()
# freq0 defined here so that Psd estimation can auto-pick if it's outside
# the range of the PSD
DEFAULT_FREQ0 = 1e-4


def simulate(
    shape=(300, 300),
    beta=8 / 3,
    resolution=60.0,
    p0=10.0,
    freq0=DEFAULT_FREQ0,
    max_amp=None,
    seed=None,
    verbose=False,
):
    """Simulate 2D isotropic turbulence with a power-law power spectral density (PSD)

    Can use either a single slope for beta, or a numpy Polynomial which is fit
    to log10(power) vs log10(frequency)

    For simple turbulence (one PSD slope), try out:
    beta = 3.0 for larger scale turbulence (long spatial correlations, large blobs)
    beta = 2.0 for middle scale turbulence
    beta = 1.3 for small scale turbulence

    Parameters
    ----------
    shape : tuple[float]
        (rows, cols), or (num_images, rows, cols) of output
        If passing 3D shape, will use same `beta` for all layers of output
        (Default value = (300, 300))
    beta : float, ndarray[float], ndarray[Polynomial], or array of polynomial coeffs,
        For scalar: power law exponent for the slope of radially averaged 2D spectrum
        Polynomial: result from fit of radially averaged log PSD vs log frequency
            (or, see result from `get_psd`)
        array of poly: one for each layer of output. Must match 3D shape.
        (Default value = 8/3)
    resolution : float
        spatial resolution of output pixels, in meters (Default value = 60.0)
    p0 : float
        Power level of PSD at the refernce frequency `freq0`
        Units are m^2 / (1/m^2) (Default value = 10.0)
    freq0 (float), reference spatial frequency where `p0` defined), in cycle / m
        (default 1e-4, or 1 cycle/10 km)
    max_amp : float or ndarray[float], optional
        maximum amplitude (in meters) of the simulated turbulence.
        Alternative to passing `p0` and `freq0`
        If passing an array, must match shape[0] of the stack
    seed : int
        number to seed random numbers, for reproducible turbulence
        (Default value = None)
    verbose : bool
        print extra debug info (Default value = False)

    Returns
    -------
    ndarray
        simulated turbulence, either 2D or 3D, depending on shape
        Units are meters

    Examples
    --------
    >>> from troposim import turbulence
    >>> out = turbulence.simulate(shape=(200, 200), beta=8/3)
    >>> # Stack of 4 images of 200x200 pixels, noise increasing in spatial scale
    >>> out = turbulence.simulate(shape=(4, 200, 200), beta=[2.0, 2.2, 2.7, 3.0])
    """
    if np.all(np.array(p0) == 0):
        return np.zeros(shape)

    try:
        num_images, length, width = shape
    except ValueError:
        num_images = 1
        length, width = shape

    # Make `beta` into an array of Polynomials, with length = num_images
    beta = _standardize_beta(beta, num_images, verbose=verbose)
    if num_images > 1:
        #  3D output, so make sure `p0` is an array of the correct length
        if np.atleast_1d(p0).size == 1:
            p0 = np.repeat(p0, num_images)
        out_list = [None for _ in range(num_images)]  # initialize list of outputs
        # https://numpy.org/devdocs/reference/random/parallel.html
        ss = SeedSequence(seed)
        child_seeds = ss.spawn(num_images)

        with ProcessPoolExecutor(max_workers=MAX_WORKERS) as executor:
            future_to_idx_map = {
                executor.submit(
                    simulate,
                    shape=(length, width),
                    beta=beta[idx],
                    p0=p0[idx],
                    freq0=freq0,
                    max_amp=max_amp,
                    resolution=resolution,
                    seed=child_seeds[idx],  # todo: how to make deterministic parallel?
                    verbose=verbose,
                ): idx
                for idx in range(num_images)
            }
            # Only show tqdm for longer simulations
            skip_bar = (length * width * num_images) < 10e6
            for fut in tqdm(
                as_completed(future_to_idx_map), total=num_images, disable=skip_bar
            ):
                idx = future_to_idx_map[fut]
                out_list[idx] = fut.result()
        return np.stack(out_list)

    # Start with the 2D PSD of white noise: flat amplitude, random phase
    rng = np.random.default_rng(seed) if seed is not None else RNG
    h = rng.uniform(size=shape)
    H = np.exp(1j * 2 * np.pi * h)

    # spatial frequencies for polynomial evaluation (units: cycles / m)
    fx = _get_freqs(width, resolution)
    fy = _get_freqs(length, resolution)
    # Broadcast 1D vectors to square: (N,1) + (1, N) = (N, N)
    f = np.sqrt(fy[:, np.newaxis] ** 2 + fx[np.newaxis, :] ** 2)

    # The power beta/2 is used because power ~ amplitude**2
    # Using amplitude means we take sqrt( k**beta) = k**(beta/2)
    beta_amp = beta / 2
    # Now evaluate each using the coefficients
    # Note: the polyval is like doing P = k ** (beta), but allows cubic, etc.
    b_coeffs = beta_amp[0].coef
    # OLD VERSION: b_coeffs = np.array([b.coef for b in beta_amp])  # for 3D eval
    # places where f=0 throw warnings in log10
    with np.errstate(invalid="ignore", divide="ignore"):
        logP = polyval(np.log10(f), b_coeffs.T)

    # create the envelope to shape the power of the white noise
    P = 10**logP
    # correct dividing by zero
    P[f == 0] = 1.0

    H_shaped = H * P
    # Make output zero mean by zeroing the top left (0 freq,constant) element
    H_shaped[0, 0] = 0.0
    out = ifft2(H_shaped, workers=-1).real
    # If passed the option of a maximum amplitude, use that and return
    return _scale_output(out, H_shaped, p0, freq0, resolution, max_amp)


def _standardize_beta(beta, num_images, verbose=False):
    """Get beta into an ndarray of Polynomials with length = num_images

    Parameters
    ----------
    beta : float or list or ndarray
        Allowed options are
        1. a single scalar, used as the linear slope
        2. a list/array of scalars, equal in length to `num_images`, treated as
                the linear slope for each image
        3. a 2d array of shape (num_images, deg+1) of polynomial coefficients
        4. a list/array of Polynomials, equal in length to `num_images`
    num_images : int
        number of images to simulate
    verbose : bool
        print extra debug info (Default value = False)

    Returns
    -------
    ndarray
        array of Polynomials, with length = num_images

    """
    # If Polynomials are passed, extract the coefficients to simplify later logic
    if isinstance(beta, Polynomial):
        beta = np.array([beta])
    if isinstance(beta, list) and isinstance(beta[0], Polynomial):
        beta = np.array([b.coef for b in beta])
    if (
        isinstance(beta, np.ndarray)
        and beta.ndim > 0
        and isinstance(beta[0], Polynomial)
    ):
        beta = np.array([b.coef for b in beta])

    # Make sure we now have just an array of coefficients
    try:
        beta = np.array(beta).astype(float)
    except TypeError:
        raise ValueError(
            f"beta must be an array of Polynomials or numeric coefficients: {beta}"
        )

    # 1. A single scalar means the linear slope used for all images
    if beta.ndim == 0:
        if beta > 0:
            # make sure the power slope is negative
            beta *= -1.0
            if verbose:
                print(f"reversed sign on scalar beta slope: {beta}")
        # Convert to linear polynomial
        beta = np.array([Polynomial([0, beta])])
        # beta = beta * np.ones(3)

    # 2. Passing a list of scalars (slopes)
    # There should be 1 for each requested image, and the input
    # should be 1D
    elif beta.ndim == 1:
        # make sure the power slope is negative for all cases
        beta[beta > 0] *= -1.0
        beta = np.array([Polynomial([0, b]) for b in beta])
    # 3. Passing an array of coefficients. Each row is the 2 (or 4)
    # coefficients of a linear (cubic) polynomial
    elif beta.ndim == 2:
        if len(beta[0]) not in (2, 4):
            raise ValueError("2D beta input must be list of polynomial coeffs")
        beta = np.array([Polynomial(b) for b in beta])
    else:
        raise ValueError(f"Invalid beta parameters: {beta}")

    if len(beta) == 1:
        beta = np.repeat(beta, num_images)
    if len(beta) != num_images:
        raise ValueError(
            f"len(beta)={len(beta)} does not match num_images={num_images}"
        )

    if verbose:
        print(f"Simulation PSD beta: {beta}")
    return beta


def _scale_output(out, H_shaped, p0, freq0, resolution, max_amp):
    # Note that this will only be called for 2D cases at bottom of `simulate`
    if max_amp is not None:
        out *= max_amp / np.max(out)
    else:
        # Otherwise, calculate the power spectral density of 1st
        # realization so that we can scale
        psd = Psd.from_image(out, resolution=resolution, freq0=freq0)
        norm_factor = np.sqrt(np.array(p0) / psd.p0)
        H_shaped *= norm_factor
        out = ifft2(H_shaped).real
    return np.squeeze(out)


def _get_freqs(N, resolution, positive=False, shift=False):
    """Get the frequencies for a given image size and resolution

    Parameters
    ----------
    N : int
        size of image
    resolution : float
        spatial resolution of image in meters
    positive : bool, optional (default=False)
        return positive frequencies
    shift : bool, optional (Default value=False)
        fftshift the frequencies to the center of the image

    Returns
    -------
    freqs : ndarray
        frequencies in cycles / m
    """
    freqs = fftfreq(N, d=resolution)
    if shift:
        freqs = fftshift(freqs)
    if positive:
        freqs = freqs[freqs > 0]
    return freqs


def _resolution_from_freqs(freqs):
    """Find the resolution back from the fft frequencies"""
    df = np.diff(freqs)[0]
    fmax = np.max(freqs) + df
    return round(1.0 / 2 / fmax)


class Psd:
    """Class for representing a power spectral density.

    Attributes
    ----------
    beta : Polynomial
        Estimated slope of PSD(s).
        Several options for input:
        1. a single scalar, used as the linear slope
        3. a 2d array of shape (1, deg+1) of polynomial coefficients
        4. a numpy.polynomial.Polynomial
        Each of these options will be converted to a Polynomial
    resolution : float
        Spatial resolution of image in meters.
    shape : tuple
        Shape of the image(s) that this PSD was calculated from.
    freq0 : Optional[float]
        Reference frequency in cycles / m, used to compute `p0`.
    freq : Optional[ndarray]
        Spatial frequencies in cycle / m.
    psd1d : ndarray
        1D power spectral density at each spatial frequency in `freq`.
        Units are m^2 / (1/m^2)
    p0 : ndarray
        Estimated power at reference frequency `freq0`.
    """

    def __init__(
        self,
        beta=None,
        resolution=None,
        shape=(300, 300),
        freq0=None,
        freq=None,
        psd1d=None,
        p0=None,
    ):
        self.beta = _standardize_beta(beta, 1)[0]
        self.shape = shape
        if resolution is None and freq is None:
            raise ValueError("Must provide freqs or resolution")
        if freq is not None:
            resolution = _resolution_from_freqs(freq)
        else:
            freq = _get_freqs(min(shape[-2:]), resolution, positive=True, shift=True)

        self.resolution = resolution
        self.freq = freq
        self.freq0 = self._get_freq0(freq, freq0)

        self.p0 = p0 if p0 is not None else self._eval_freq(self.freq0, self.beta)
        if psd1d is not None:
            self.psd1d = np.array(psd1d)
        else:
            psd1d = self._eval_freq(self.freq, self.beta)
            psd1d *= self.p0 / self._eval_freq(self.freq0, self.beta)
            self.psd1d = psd1d

    @staticmethod
    def _eval_freq(freq, beta):
        """Evaluate the power spectral density at a given frequency"""
        return np.power(10, beta(np.log10(freq)))

    def simulate(self, shape=None, **kwargs):
        """Simulate a power spectral density.
        See docstring of turbulence.simulate for full information
        """
        if shape is None:
            shape = self.shape
        return simulate(
            shape=shape,
            beta=self.beta,
            p0=self.p0,
            freq0=self.freq0,
            resolution=self.resolution,
            **kwargs,
        )

    @classmethod
    def from_image(
        cls,
        image,
        resolution,
        freq0=None,
        deg=3,
        crop=True,
        N=None,
    ):
        """Get the radially averaged 1D PSD of in input 2D/3D image

        Parameters
        ----------
        image : 2D or 3D ndarray
            image or stack of images, units in meters
        resolution : float
            spatial resolution of input image in meters (Default value = 60)
        freq0 : float
            Reference spatial frequency in cycle / m.
            If not passed, default value = 1e-4, as long as it falls
            within the possible range of the image.
            If the image is too small, will divide by 2 until it fits.
        deg : int
            degree of Polynomial to fit to PSD. default = 3, cubic
        crop : bool
            crop the image into a square image with fewest non-zero pixels
            (Default value = True)
        N : int
            Crop the image to a square with max size N.
            If None, will use the largest size that keeps the image square.

        Returns
        -------
        Psd object, or PsdStack if image is 3D

        Examples
        --------
        >>> from troposim.turbulence import Psd
        >>> import numpy as np
        >>> image = np.random.rand(100, 100)
        >>> psd = Psd.from_image(image, resolution=180)
        """
        if image.ndim > 2:
            return PsdStack.from_images(
                image,
                resolution=resolution,
                freq0=freq0,
                deg=deg,
                crop=crop,
                N=N,
            )

        psd2d = cls._get_psd2d(image, shift=True, crop=crop, N=N, resolution=resolution)

        # calculate the radially average spectrum
        # freq, psd1d = radial_average_spectrum(psd2d, resolution)
        freq, psd1d = cls._average_radially(psd2d, resolution=resolution)
        freq0 = cls._get_freq0(freq, freq0)

        # calculate slopes from spectrum
        beta_hat = cls.fit_psd1d(freq, psd1d, deg=deg)
        p0 = cls._eval_freq(freq0, beta_hat)
        beta_hat = np.array([beta_hat])
        return Psd(
            beta=beta_hat,
            resolution=resolution,
            shape=image.shape,
            psd1d=psd1d,
            freq0=freq0,
            p0=p0,
        )

    @staticmethod
    def _get_freq0(freq, freq0):
        """Check that freq0 is in the range of the image"""
        # If it's specified and incorrect, raise an error
        if freq0 is not None:
            if freq0 < np.min(freq) or freq0 > np.max(freq):
                raise ValueError(
                    f"freq0={freq0} is out of range {np.min(freq):.2E}-{np.max(freq):.2E}."
                )
            else:
                return freq0
        # Otherwise, pick one for the user
        # print("Finding suitable reference frequency")
        if freq0 is None:
            freq0 = DEFAULT_FREQ0
        # For smaller images, 1e-4 might be too small, so try increasing
        pos_freqs = freq[freq > 0]
        while freq0 < np.min(pos_freqs):
            freq0 *= 2
        # and double check that we didn't overshoot somehow
        freq0 = np.clip(freq0, np.min(pos_freqs), np.max(pos_freqs))
        return freq0

    def save(self, filename="psd_params.npz", save_dir=None):
        """Save the PSD parameters to a file.

        Parameters
        ----------
        filename : str, optional
            Filename to save the PSD parameters
            If not passed, will save to `psd_params.npz`
        save_dir : str, optional
            Directory to save the PSD parameters
        """
        if filename is None:
            filename = "psd_params.npz"
        if save_dir is None:
            save_dir = Path()
        save_name = save_dir / Path(filename)

        np.savez(
            save_name,
            beta=[b.coef],
            psd1d=self.psd1d,
            resolution=self.resolution,
            shape=self.shape,
            freq0=self.freq0,
        )
        return save_name

    @classmethod
    def load(cls, filename):
        """Load a saved PSD parameter file

        Parameters
        ----------
            filename (str or Path): Name of the file to load.

        Returns
        -------
            new Psd object
        """
        # TODO: Do i also wanna allow other formats (h5py)?
        if not str(filename).endswith(".npz"):
            filename = str(filename) + ".npz"
        if not Path(filename).exists():
            raise ValueError(f"PSD parameter from {filename} does not exist")

        with np.load(filename) as data:
            beta = data["beta"]
            shape = data.get["shape"]
            resolution = data["resolution"]
            freq0 = data.get("freq0")
            psd1d = data.get("psd1d")

        return cls(
            beta=beta, resolution=resolution, shape=shape, freq0=freq0, psd1d=psd1d
        )

    def asdict(self, include_psd1d=True) -> dict:
        """Save the PSD parameters to a dictionary.

        Returns
        -------
        dict
            Dictionary representation of the PSD parameters
        """
        d = {
            # "p0": self.p0.item(),
            "beta": [self.beta.coef.tolist()],  # list of lists, so it's (1, 4) not (4,)
            "shape": self.shape,
            "resolution": self.resolution,
            "freq0": self.freq0,
            # "psd1d": self.psd1d.ravel().tolist(),
            # Freq can be found from the resolution, so don't save it
            # "freq": self.freq.tolist(),
        }
        if include_psd1d and self.psd1d is not None:
            d["psd1d"] = self.psd1d.ravel().tolist()
        return d

    @classmethod
    def from_dict(cls, psd_dict: dict):
        """Load a saved PSD parameter file

        Parameters
        ----------
            psd_dict (dict): Dictionary representation of the PSD parameters

        Returns
        -------
            new Psd object
        """
        beta = _standardize_beta(psd_dict["beta"], 1)[0]
        shape = tuple(psd_dict["shape"])
        resolution = psd_dict["resolution"]
        freq0 = psd_dict.get("freq0")
        psd1d = psd_dict.get("psd1d")
        return cls(
            beta=beta, shape=shape, resolution=resolution, freq0=freq0, psd1d=psd1d
        )

    @staticmethod
    def _get_psd2d(
        image=None,
        shift=True,
        crop=False,
        N=None,
        resolution=60.0,
    ):
        """Calculate the 2d Power spectral density of and image/stack of images

        Parameters
        ----------
        image : ndarray
            image or stack of images (Default value = None)
        shift : bool
            apply `fftshift` in the Fourier domain. Defaults to True.
        crop : bool
            Crop the image to a square of size `N`. Defaults to False.
        N : [type]
            For cropping, size of square. Defaults to None.
            If None, will find the largest sqaure with greatest number of nonzeros
            (see `utils.crop_square_data`)
        resolution : float
            spatial resolution of input image in meters
            used if to normalize.

        Returns
        -------
        psd2d: ndarray
            2D Power spectral density of input image.
        """
        # use the square part of the matrix for spectrum calculation
        if crop:
            image = utils.crop_square_data(image, N=N)
        rows, cols = image.shape[-2:]

        # calculate the normalized power spectrum (spectral density)
        fdata2d = fft2(np.nan_to_num(image, nan=0.0))
        if shift:
            fdata2d = fftshift(fdata2d, axes=(-2, -1))
        psd2d = np.abs(fdata2d) ** 2 / (rows * cols)

        # Convert to density units (m^2 / (1/m^2), or (Amplitude^2) / (1 / (sampling units)^2))
        # Here the sampling units are meters, given by the resoltution
        psd2d *= resolution**2
        # print(f"mult psd2d by {resolution**2:.1f}")
        # print(f"if in [km]: {(resolution/1000)**2:.1f}")
        # print(f"So dividing by Fs**2 leads to boost of {(1000/resolution)**2:.1f}")

        return psd2d

    @staticmethod
    def _average_radially(
        psd2d,
        resolution=60.0,
    ):
        """Calculate the radially averaged power spectrum (assumes isotropy)

        Parameters
        ----------
        psd2d : ndarray
            precomputed 2d psd
        resolution : float
            (Default value = 60.0)

        Returns
        -------
        freq_pos : ndarray
            radial frequency in cycles / m
        psd1d : ndarray
            power spectral density in (m^2) / (1 / m^2)
        """
        h, w = psd2d.shape[-2:]
        hc, wc = h // 2, w // 2
        # Only extend out to shortest dimension
        # this will miss power contributions in 'corners' r > min(hc, wc)
        num_r = min(hc, wc)

        # create an array of integer radial distances from the center
        Y, X = np.ogrid[0:h, 0:w]
        # These act as "labels" in the ndimage summation
        r = np.hypot(X - wc, Y - hc).round().astype(int)
        # r *= resolution # for the actual radial distances in image

        # average all psd2d pixels with label 'r' for 0<=r<=wc
        psd1d = ndimage.mean(psd2d, r, index=np.arange(0, num_r))

        # Also return the positive frequencies with this
        N = min(h, w)
        freq_pos = _get_freqs(N, resolution, positive=True, shift=True)
        # For even-sized images, there's an offset
        # see fftfreq's case, where there are (N - 1) // 2 positive freqs
        if N % 2 == 0:
            psd1d = psd1d[1:]
        return freq_pos, psd1d

    @staticmethod
    def fit_psd1d(freq, psd, deg=3, verbose=False):
        """Fit the slope `beta` and p0 of a 1D PSD in loglog scale
        p = p0 * (freq/freq0)^(-beta)

        Python translation of pslope.m (Ramon Hanssen, 2000), with added
        higher order polynomial option

        Parameters
        ----------
        freq : 1D / 2D ndarray
            in cycle / m
        psd : 1D / 2D ndarray
            power spectral density
        freq0 : float
            reference frequency in cycle / m
        deg : int
            degree of polynomial fit to log-log plot (default = 3, cubic)
        verbose : bool
            (Default value = False)

        Returns
        -------
        p0 : float
            power at reference frequency
        beta : np.Polynomial
            polynomial representing the fitted 1D PSD
        """
        freq = freq.flatten()
        psd = psd.flatten()

        # check if there is zero frequency. If yes, remove it.
        idx = freq != 0.0
        if not np.all(idx):
            freq = freq[idx]
            psd = psd[idx]

        if np.all(psd == 0.0):
            print("Warning: All PSD values are zero")
            beta_poly = Polynomial.fit(freq, psd, deg=deg, domain=[])
            return 0.0, beta_poly

        # convert to log-log scale
        logf = np.log10(freq)
        logp = np.log10(psd)

        # Ignore once it's flat at basically 0
        nonzero_idx = (np.abs(logp) > 1e-16) & np.isfinite(logp) & ~np.isnan(logp)
        logp = logp[nonzero_idx]
        logf = logf[nonzero_idx]

        # Note: domain=[] necessary to return expected typical power series coefficients
        # Makes it equivalent to older `np.polyfit(logf, logp, deg=deg)`
        # See https://github.com/numpy/numpy/issues/9533, https://stackoverflow.com/a/52339988
        # Reminder that fitting without domain=[] requires .convert() to remove affine map
        beta_poly = Polynomial.fit(logf, logp, deg=deg, domain=[])
        if verbose:
            print(f"Polyfit fit: {beta_poly}")

        return beta_poly

    @classmethod
    def from_hdf5(
        cls,
        hdf5_file,
        dataset,
        resolution,
        freq0=None,
        deg=3,
        crop=True,
    ):
        """Find the PSD of a stack of images saved in an HDF5 file

        Parameters
        ----------
        h5_file : str
            Name of HDF5 file
        dataset : str
            dataset within the HDF5 containing the images
        resolution : float

        Returns
        -------
        Psd object
        """
        import h5py

        with h5py.File(hdf5_file, "r") as f:
            dset = f[dataset]
            if dset.ndim == 3 and len(dset) > 1:
                return PsdStack.from_images(
                    dset, resolution, freq0=freq0, deg=deg, crop=crop
                )
            else:
                return cls.from_image(dset, resolution, freq0=freq0, deg=deg, crop=crop)
            # stack = f[dataset][:]

    def plot(self, idxs=None, ax=None, **kwargs):
        from troposim import plotting

        if idxs is None:
            idxs = slice(None)
        if "color" not in kwargs:
            kwargs["color"] = "black"

        return plotting.plot_psd1d(self.freq, self.psd1d[idxs].T, ax=ax, **kwargs)

    def __repr__(self):
        # The repr version can be used to reconstruct the object
        return self._to_str(poly_as_str=False)

    def __str__(self):
        return self._to_str(poly_as_str=True, precision=2)

    def _to_str(self, poly_as_str: bool, precision=None):
        with np.printoptions(precision=precision):
            if poly_as_str:
                beta_str = str(self.beta)
            else:
                beta_str = repr(self.beta)
        s = f"Psd(beta={beta_str}, shape={self.shape}, resolution={self.resolution}"
        if self.freq0 is not None:
            s += f", freq0={self.freq0}"
        s += ")"
        return s

    def __rich_repr__(self):
        yield "beta", self.beta
        yield "shape", self.shape
        yield "resolution", self.resolution
        yield "freq0", self.freq0

    def __eq__(self, other):
        a = self.shape == other.shape
        b = np.array_equal(self.beta, other.beta)
        c = self.resolution == other.resolution
        return a and b and c

    def __add__(self, other):
        self._assert_compatible(other)
        return Psd(
            beta=(self.beta + other.beta) / 2,  # average the beta polynomials
            resolution=self.resolution,
            shape=self.shape,
            freq0=self.freq0,
            psd1d=self.psd1d + other.psd1d,
        )

    def __mul__(self, c):
        if isinstance(c, (int, float)):
            # Scalar multiplication
            # Note: this is designed to match the result of multiplying the
            # 1D PSD by a constant. The 1D PSD is a log-log curve, the beta
            # polynomial gets changed 1) only in the constant term, and 2) by
            # a factor of log10(c). The 1D PSD is then multiplied by c.
            psd1d = self.psd1d * c if self.psd1d is not None else None
            constant_factor = np.log10(c)
            beta = self.beta.copy()
            beta.coef[0] += constant_factor
            return Psd(
                beta=beta,
                resolution=self.resolution,
                shape=self.shape,
                freq0=self.freq0,
                psd1d=psd1d,
            )
        else:
            raise TypeError("Multiplication only supported by scalar")

    def __rmul__(self, c):
        return self.__mul__(c)

    # the div allows us to do (psd1 + psd2) / 2
    def __truediv__(self, c):
        if isinstance(c, (int, float)):
            # See comment in __mul__ for explanation of this
            psd1d = self.psd1d * c if self.psd1d is not None else None
            constant_factor = np.log10(c)
            beta = self.beta.copy()
            beta.coef[0] -= constant_factor
            return Psd(
                beta=beta,
                freq0=self.freq0,
                shape=self.shape,
                resolution=self.resolution,
                psd1d=psd1d,
            )
        else:
            raise TypeError("Division only supported by scalar")

    def _assert_compatible(self, other):
        if not isinstance(other, Psd):
            raise TypeError("Both objects must be `Psd` instances")
        if not self.freq0 == other.freq0:
            raise ValueError("Psd objects must have same freq0")
        if not self.shape == other.shape:
            raise ValueError("Psd objects must have same shape")

    def copy(self, deep=True):
        return copy.deepcopy(self) if deep else copy.copy(self)

    @classmethod
    def from_p0_beta(cls, p0, beta, resolution, shape, freq0):
        """Reconstruct 1D power spectral density array from p0 and beta

        Parameters
        ----------
        p0 : float
            power spectral density in m^2 at `f0` Hz
        beta
            power spectra slope in loglog scale
            if scalar passed, assuming it single slope of loglog plot
        resolution : float
            spatial resolution of the image in meters
        shape : tuple[int]
            desired shape of image.
        freq0 : float
            reference spatial frequency in cycle / m

        Returns
        -------
        Psd object
        """
        # frequency for x-axis after FFT
        N = min(shape[-2:])
        freq = _get_freqs(N, resolution, positive=True, shift=True)
        freq0 = cls._get_freq0(freq, freq0)

        # Make `beta` into a Polynomial
        beta = _standardize_beta(beta, 1)[0]

        psd1d = cls._eval_freq(freq, beta)
        psd1d *= p0 / cls._eval_freq(freq0, beta)

        beta_hat = cls.fit_psd1d(freq, psd1d, deg=beta.degree())
        return Psd(
            beta=beta_hat,
            resolution=resolution,
            shape=tuple(shape),
            freq0=freq0,
            # Note: not including psd1d here because it came from beta evaluation,
            # not from fitting.
        )


class PsdStack:
    psd_list: List[Psd]

    @classmethod
    def from_images(
        cls,
        stack,
        resolution,
        freq0=None,
        deg=3,
        crop=True,
        N=None,
    ):
        """Find the PSD estimates for a stack of images.

        Parameters
        ----------
        stack : 3D ndarray
            displacement in meters (num_images, rows, cols)
        resolution : float
            spatial resolution of input data in meters
        freq0 : float
            reference spatial frequency in cycle / m.
        deg : int
            degree of Polynomial to fit to PSD. default = 3, cubic
        crop : bool
            crop the data into a square image with fewest non-zero pixels
            (Default value = True)
        N : int
            size to crop square (defaults to size of smaller side)
        resolution : float
            (Default value = 60.0)
        freq0 : float
            (Default value = 1e-4)

        Returns
        -------
        Psd object with iterables for p0, beta, and psd1d
        """
        psd_list = []
        skip_progress_bar = len(stack) < 5
        for image in tqdm(stack, disable=skip_progress_bar):
            psd_list.append(
                Psd.from_image(
                    image,
                    resolution=resolution,
                    freq0=freq0,
                    deg=deg,
                    crop=crop,
                    N=N,
                )
            )
        return cls(psd_list)

    def simulate(
        self, num_days=None, shape=None, randomize=False, **kwargs
    ) -> np.ndarray:
        """Simulate the stack of images from the PSD

        Returns
        -------
        ndarray
            simulated stack of images
        """
        if num_days is None:
            num_days = self.shape[0]
        if shape is None:
            shape = self.shape[-2:]
        stack_shape = (num_days,) + shape
        if not randomize:
            # if len(beta)=3 does not match num_images=50
            if len(self.beta) > 1 and len(self.beta) != num_days:
                raise ValueError(
                    "Number of days must match number of beta values if randomize=False"
                )
            return simulate(
                shape=stack_shape,
                beta=self.beta,
                resolution=self.resolution,
                p0=self.p0,
                freq0=self.freq0,
                **kwargs,
            )

        # Now we need to randomize the PSDs with replacement
        psd_list = self.psd_list
        # if num_days > len(psd_list):
        #     # pad the list with copies of itself if we need more
        #     psd_list = psd_list * (num_days // len(psd_list) + 1)

        psd_list = np.random.choice(psd_list, size=(num_days,), replace=True)

        # Get a set of random indices with replacement
        beta = np.array([psd.beta for psd in psd_list]).ravel()
        p0_array = np.array([psd.p0 for psd in psd_list]).ravel()

        return simulate(
            shape=stack_shape,
            beta=beta,
            resolution=self.resolution,
            p0=p0_array,
            freq0=self.freq0,
            **kwargs,
        )

    def __init__(self, psd_list):
        if not all([psd.shape == psd_list[0].shape for psd in psd_list]):
            raise ValueError("All Psd objects must have same shape")
        if not all([psd.resolution == psd_list[0].resolution for psd in psd_list]):
            raise ValueError("All Psd objects must have same resolution")
        if not all([psd.freq0 == psd_list[0].freq0 for psd in psd_list]):
            raise ValueError("All Psd objects must have same freq0")
        self.psd_list = psd_list

    def __str__(self):
        return f"PsdStack with {len(self)} images"

    def __repr__(self) -> str:
        return (
            f"PsdStack(psd_list=[{', '.join([repr(psd) for psd in self.psd_list])}]])"
        )

    def __rich_repr__(self):
        yield "psd_list", self.psd_list[:5]
        if len(self.psd_list) > 5:
            yield f"...({len(self.psd_list) - 5} more)"

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, PsdStack):
            return False
        return all([psd1 == psd2 for psd1, psd2 in zip(self.psd_list, o.psd_list)])

    @property
    def p0(self):
        return np.array([psd.p0 for psd in self.psd_list])

    @property
    def beta(self):
        return np.array([psd.beta for psd in self.psd_list])

    @property
    def psd1d(self):
        return np.array([psd.psd1d for psd in self.psd_list])

    @property
    def freq(self):
        return self.psd_list[0].freq

    @property
    def freq0(self):
        return self.psd_list[0].freq0

    @property
    def resolution(self):
        return self.psd_list[0].resolution

    @property
    def shape(self):
        return (len(self.psd_list),) + self.psd_list[0].shape

    def __getitem__(self, idx):
        p = self.psd_list[idx]
        if isinstance(idx, int):
            return p
        return PsdStack(p)

    def __setitem__(self, idx, value):
        self.psd_list[idx] = value

    def __len__(self):
        return len(self.psd_list)

    def __add__(self, other):
        if not isinstance(other, PsdStack):
            raise TypeError("Both objects must be `PsdStack` instances")
        return PsdStack(self.psd_list + other.psd_list)

    def asdict(self, include_psd1d=True):
        return dict(psd_list=[psd.asdict(include_psd1d) for psd in self.psd_list])

    @classmethod
    def from_dict(cls, d):
        return cls([Psd.from_dict(psd) for psd in d["psd_list"]])

    def plot(self, idxs=None, ax=None, **kwargs):
        from troposim import plotting

        if idxs is None:
            idxs = slice(None)
        if "color" not in kwargs:
            kwargs["color"] = "grey" if len(self) > 1 else "black"

        return plotting.plot_psd1d(self.freq, self.psd1d[idxs].T, ax=ax, **kwargs)



"""
This module exports a Log class that wraps the logging python package

Uses the standard python logging utilities, just provides
nice formatting out of the box.

Usage:

    from .log import get_log
    logger = get_log()

    logger.info("Something happened")
    logger.warning("Something concerning happened")
    logger.error("Something bad happened")
    logger.critical("Something just awful happened")
    logger.debug("Extra printing we often don't need to see.")
    # Custom output for this module:
    logger.success("Something great happened: highlight this success")
"""
import logging
import time
from logging import Formatter

COLORS = False


def get_log(name=__file__, debug=False, verbose=False):
    """Creates a nice log format for use across multiple files.
    
    Default logging level is INFO

    Parameters
    ----------
    name : Optional[str]
        The name the logger will use when printing statements (Default value = __file__)
    debug : Optional[bool]
        If true, sets logging level to DEBUG (Default value = False)
    verbose :
        (Default value = False)

    Returns
    -------

    
    """
    logger = logging.getLogger(name)
    return format_log(logger, debug=debug, verbose=verbose)


def format_log(logger, debug=False, verbose=False):
    """Makes the logging output pretty and colored with times

    Parameters
    ----------
    logger : logging.Logger
        
    debug : bool
        (Default value = False)
    verbose : bool
        (Default value = False)

    Returns
    -------
    logging.Logger
    """
    log_level = logging.DEBUG if debug else logging.INFO
    format_ = "[%(asctime)s] [%(levelname)s %(filename)s] %(message)s"
    formatter = Formatter(format_, datefmt="%m/%d %H:%M:%S")

    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    logging.SUCCESS = 25  # between WARNING and INFO
    logging.addLevelName(logging.SUCCESS, "SUCCESS")
    setattr(
        logger,
        "success",
        lambda message, *args: logger._log(logging.SUCCESS, message, args),
    )

    if not logger.handlers:
        logger.addHandler(handler)
        logger.setLevel(log_level)

        if verbose:
            logger.info("Logger initialized: %s" % (logger.name,))

    if debug:
        logger.setLevel(debug)

    return logger


logger = get_log()


def log_runtime(f):
    """Decorator to log how long a decorated function takes to run

    Usage
    -----
    @log_runtime
    def test_func():
        return 2 + 4
    """

    def wrapper(*args, **kwargs):
        """

        Parameters
        ----------
        *args :
            
        **kwargs :
            

        Returns
        -------

        
        """
        t1 = time.time()

        result = f(*args, **kwargs)

        t2 = time.time()
        elapsed_time = t2 - t1
        time_string = "Total elapsed time for {} : {} minutes ({} seconds)".format(
            f.__name__,
            "{0:.2f}".format(elapsed_time / 60.0),
            "{0:.2f}".format(elapsed_time),
        )

        logger.info(time_string)
        return result

    return wrapper



"""Simulate stratified atmospheric delay

Based mostly on Bekaert, 2015:
A spatially variable power law tropospheric correction technique for InSAR data
"""
import numpy as np


def simulate(dem, K_params={}, h0_params={}, alpha_params={}):
    """Simulate a set of stratified delay maps from a DEM
    Uses the power law scaling from Bekaert, 2015:
        `K * (h0 - h) ** alpha`

    Parameters
    ----------
    dem : ndarray
        DEM heights (in meters)
    K_params : dict
        parameters for sampling the atmospheric constant:
        options:
        'mean' (default = 0)
        'sigma': (default = 5e-6)
        'shape': (default = (1,))
        K is sampled from a laplacian distribution
        units: radians / meters^alpha
    h0_params : float
        top of atmosphere height.
        options:
        'mean' (default = 7000)
        'sigma': (default = 100)
        'shape': (default = (1,))
    alpha_params :
        parameters for sampling the power law scaling:
        options:
        'mean' (default = 1.4)
        'sigma': (default = 0.1)
        'shape': (default = (1,))

    Returns
    -------
    np.ndarray
    """
    K = sample_K(**K_params)
    alpha = sample_alpha(**alpha_params)
    h0 = sample_h0(**h0_params)
    return make_stratified_delay(dem, K, h0=h0, alpha=alpha)


def make_stratified_delay(dem, K, h0=7000, alpha=1.4, zero_mean=True):
    """Create a stratified delay map from a DEM

    Parameters
    ----------
    dem : ndarray
        DEM heights (in meters)
    K : float
        atmospheric constant: `K (h0 - h) ** alpha`
        units: radians / meters^alpha
    h0 : float
        top of atmosphere height. Defaults to 7000.
    alpha : float
        power law scaling. Defaults to 1.4.
    zero_mean :
        (Default value = True)

    Returns
    -------
    ndarray
    """
    K, h0, alpha = _get_param_sizes(K, h0, alpha)

    if np.any(h0 < dem.max()):
        raise ValueError(f"h0 = {h0} must be higher than DEM")
    dh = h0 - dem
    delay = K * dh**alpha
    if zero_mean:
        if delay.ndim == 3:
            delay -= delay.mean(axis=(1, 2), keepdims=True)
        else:
            delay -= delay.mean()
    return delay


def sample_K(mean=0.0, sigma=5e-6, shape=(1,)):
    """Sample a random K value from a normal distribution

    Parameters
    ----------
    mean :
        (Default value = 0.0)
    sigma :
        (Default value = 5e-6)
    shape : in or tuple[int]
        (Default value = (1,))

    Returns
    -------
    np.ndarray
    """
    return np.random.laplace(loc=mean, scale=sigma, size=shape)


def sample_alpha(mean=1.4, sigma=0.1, shape=(1,)):
    """Sample a random alpha value from a normal distribution

    Parameters
    ----------
    mean :
        (Default value = 1.4)
    sigma :
        (Default value = 0.1)
    shape :
        (Default value = (1, ))
    """
    return np.random.normal(loc=mean, scale=sigma, size=shape)


def sample_h0(mean=7000, sigma=100, shape=(1,)):
    """Sample a random h0 value from a normal distribution

    Parameters
    ----------
    mean :
        (Default value = 7000)
    sigma :
        (Default value = 100)
    shape :
        (Default value = (1, ))
    """
    return np.random.normal(loc=mean, scale=sigma, size=shape)


def _get_param_sizes(K, h0, alpha):
    if np.isscalar(K):
        K = np.array(K).reshape((1, 1, 1))
    else:
        K = np.array(K).reshape((len(K), 1, 1))
    if np.isscalar(h0):
        h0 = np.array(h0).reshape((1, 1, 1))
    else:
        h0 = np.array(h0).reshape((len(h0), 1, 1))
    if np.isscalar(alpha):
        alpha = np.array(alpha).reshape((1, 1, 1))
    else:
        alpha = np.array(alpha).reshape((len(alpha), 1, 1))
    nk, nh, na = len(K), len(h0), len(alpha)
    lengths = set([nk, nh, na])
    if len(lengths) > 2 or (len(lengths) > 1 and 1 not in lengths):
        raise ValueError("K, h0, and alpha must all have the same length or length 1")
    return K, h0, alpha



import numpy as np
import pytest
from numpy.polynomial import Polynomial
from numpy.testing import assert_allclose

from troposim import turbulence


def test_default():
    turbulence.simulate()


def test_shapes():
    freq0 = 2.0e-3
    out = turbulence.simulate(shape=(10, 10), freq0=freq0)
    assert out.shape == (10, 10)
    out = turbulence.simulate(shape=(2, 10, 10), freq0=freq0)
    assert out.shape == (2, 10, 10)
    out_rect = turbulence.simulate(shape=(3, 10, 20), freq0=freq0)
    assert out_rect.shape == (3, 10, 20)


def test_get_psd():
    """ """
    shape = (200, 200)
    p0 = 1e-3
    b = -2.5
    freq0 = 2.0e-3
    out = turbulence.simulate(shape=shape, beta=b, freq0=freq0, p0=p0)
    psd = turbulence.Psd.from_image(out, freq0=freq0, deg=1)

    beta_tol = 0.15
    assert abs(psd.beta.item().coef[1] - b) < beta_tol
    p0_tol = 0.05
    assert abs(psd.p0 - p0) < p0_tol


def test_zero_psd():
    """ """
    shape = (200, 200)
    turbulence.Psd.from_image(np.zeros(shape), deg=1)
    turbulence.Psd.from_image(np.zeros(shape), deg=3)


def _beta_is_valid(beta, num_images):
    beta = turbulence._standardize_beta(beta, num_images)
    return (len(beta) == num_images) and isinstance(beta[0], Polynomial)


def test_standardize_beta():
    assert _beta_is_valid(2.5, 1)
    assert _beta_is_valid(2.5, 3)
    assert _beta_is_valid(Polynomial([0, 2.5]), 1)

    assert _beta_is_valid([2.0, 2.5], 2)
    assert _beta_is_valid(np.array([Polynomial([0, 2.5]), Polynomial([0, 2.5])]), 2)

    with pytest.raises(ValueError):
        # Wrong number of images requested
        _beta_is_valid([2.0, 2.5, 3.0, 3.0, 5], 2)

    # Check that we can pass iterables of coefficients
    assert _beta_is_valid([[0.0, 2.0], [0, 2.5]], 2)
    # Test cubic polys too
    assert _beta_is_valid([[0.0, 1.0, 2.0, 3.0]], 2)
    assert _beta_is_valid([[0.0, 1.0, 2.0, 3.0], [0.0, 1.0, 2.0, 3.0]], 2)

    with pytest.raises(ValueError):
        # Wrong number of images requested
        assert _beta_is_valid([[0.0, 2.0], [0, 2.5]], 1)
        assert _beta_is_valid([[0.0, 2.0], [0, 2.5]], 3)


def test_beta_types():
    """ """
    # Scalar test
    b = 2.5
    beta_tol = 0.15  # For comparing two way simulate
    freq0 = 2.0e-3
    shape2d = (200, 200)
    out = turbulence.simulate(shape=shape2d, beta=b, freq0=freq0)
    assert out.shape == shape2d

    shape3d = (4, 200, 200)
    out = turbulence.simulate(shape=shape3d, beta=b, freq0=freq0)
    assert out.shape == shape3d
    psd = turbulence.Psd.from_image(out, freq0=freq0, deg=1)
    assert len(psd.beta) == shape3d[0]

    # Note that the slope will always be negative,
    # even though we allow user to pass positive sloeps
    bslope = -b
    for bh in psd.beta:
        assert abs(bh.coef[1] - bslope) < beta_tol

    # Same thing, but manually specifying same b
    out = turbulence.simulate(shape=shape3d, beta=[b, b, b, b], freq0=freq0)
    psd = turbulence.Psd.from_image(out, freq0=freq0, deg=1)
    for bh in psd.beta:
        assert abs(bh.coef[1] - bslope) < beta_tol

    beta_list = [-2.0, -2.3, -2.7, -2.9]
    out = turbulence.simulate(shape=shape3d, beta=beta_list, freq0=freq0)
    psd = turbulence.Psd.from_image(out, freq0=freq0, deg=1)
    for bh, b in zip(psd.beta, beta_list):
        assert abs(bh.coef[1] - b) < beta_tol


def test_load_save_psd(tmp_path):
    fname = tmp_path / "test_psd.npz"
    # Scalar test
    b = 2.5
    freq0 = 2.0e-3
    shape2d = (200, 200)
    out = turbulence.simulate(shape=shape2d, beta=b, freq0=freq0)
    psd = turbulence.Psd.from_image(out, freq0=freq0)
    psd.save(fname)
    psd2 = turbulence.Psd.load(fname)
    assert psd == psd2

    shape3d = (4, 200, 200)
    beta_list = [-2.0, -2.3, -2.7, -2.9]
    fname3d = tmp_path / "test_psd_3d.npz"

    out = turbulence.simulate(shape=shape3d, beta=beta_list, freq0=freq0)

    # p0, beta, f, psd = turbulence.get_psd(out, freq0=freq0, deg=3, filename=fname3d)
    # p0_2, beta_2, f_2, psd_2 = turbulence.get_psd(
    #     fname, freq0=freq0, deg=1, filename=fname3d
    # )


def test_from_hdf5(tmp_path):
    import h5py

    noise4 = turbulence.simulate(shape=(4, 200, 200))
    fname = tmp_path / "test_psd.h5"
    dset = "noise"
    with h5py.File(fname, "w") as hf:
        hf.create_dataset(dset, data=noise4)
    psd4 = turbulence.Psd.from_hdf5(fname, dset, resolution=60)
    assert psd4.p0.shape == (4,)
    assert psd4.beta.shape == (4,)
    assert psd4.freq.shape == (99,)
    assert psd4.psd1d.shape == (4, 99)


def test_max_amp():
    max_amp = 0.05  # meters
    noise = turbulence.simulate(shape=(100, 100), max_amp=max_amp)
    assert_allclose(np.max(noise), max_amp)
    assert noise.shape == (100, 100)

    noise3d = turbulence.simulate(shape=(3, 100, 100), max_amp=max_amp)
    assert noise3d.shape == (3, 100, 100)
    assert_allclose(np.max(noise3d, axis=(1, 2)), max_amp)


def test_resolution_freq_change():
    N = 200
    res = 60
    assert res == turbulence._resolution_from_freqs(turbulence._get_freqs(N, res))


def test_get_auto_freq0():
    res = 100
    noise = turbulence.simulate(shape=(200, 200), resolution=res)
    psd = turbulence.Psd.from_image(noise, resolution=res)

    assert psd._get_freq0(psd.freq, None) == 1e-4
    assert psd._get_freq0(psd.freq, 1e-4) == 1e-4
    assert psd._get_freq0(psd.freq, 1e-3) == 1e-3
    with pytest.raises(ValueError):
        assert psd._get_freq0(psd.freq, 1e-6)  # too small
        assert psd._get_freq0(psd.freq, 1)  # too big

    # Check the auto-picking for a small image
    noise = turbulence.simulate(shape=(50, 50), resolution=res, p0=1, freq0=1e-3)
    psd = turbulence.Psd.from_image(noise, resolution=res)
    assert psd._get_freq0(psd.freq, None) == 2e-4
    psd = turbulence.Psd.from_image(noise[:25, :25], resolution=res)
    assert psd._get_freq0(psd.freq, None) == 4e-4


def test_append_psds():
    noise = turbulence.simulate(shape=(200, 200))
    noise2 = turbulence.simulate(shape=(200, 200))

    psd = turbulence.Psd.from_image(noise, resolution=60)
    psd2 = turbulence.Psd.from_image(noise2, resolution=60)
    assert len(psd) == len(psd2) == 1
    psd.append(psd2)
    assert len(psd) == 2
    psd.append(psd)
    assert len(psd) == 4

    assert psd.p0.shape == (4,)
    assert psd.beta.shape == (4,)
    assert psd.freq.shape == (99,)
    assert psd.psd1d.shape == (4, 99)

    psd_wrongshape = turbulence.Psd.from_image(turbulence.simulate(shape=(250, 250)))
    with pytest.raises(ValueError):
        psd.append(psd_wrongshape)


@pytest.fixture
def psd():
    return turbulence.Psd.from_image(turbulence.simulate(shape=(200, 200)))


@pytest.fixture
def psd5():
    return turbulence.Psd.from_image(turbulence.simulate(shape=(5, 200, 200)))


def test_div(psd):
    psd == ((psd + psd) / 2)


def test_len_psds(psd5):
    assert len(psd5) == 5


def test_dict_roundtrip(psd5):
    psd_dict = psd5.asdict()
    psd5b = turbulence.Psd.from_dict(psd_dict)
    assert psd5 == psd5b


# TODO: something weird with a 0 one
def test_zero_img():
    out = turbulence.simulate(
        shape=(3, 100, 100), beta=[2.0, 2.7, 3.0], verbose=True, freq0=1e-3
    )
    out[1] = 0
    psd = turbulence.Psd.from_image(out)
    psd.simulate()



from dataclasses import dataclass
from datetime import date, timedelta
from itertools import combinations
from typing import Optional, Tuple

import numpy as np

from . import turbulence, utils
from .log import get_log


logger = get_log(__name__)


@dataclass
class IgramMaker:
    """Class for creating synthetic interferograms"""

    psd_stack: turbulence.PsdStack
    num_days: int = 10
    shape: Optional[Tuple[int, int]] = None
    randomize: bool = True
    # resolution: int = 400
    # p0_default: float = 10.0
    # freq0: float = 1e-4
    start_date: date = date(2018, 1, 1)
    repeat_interval_days: int = 12
    sar_stack: np.ndarray | None = None
    to_cm: bool = True
    defo_rates: Tuple[float] = (0.0,)
    defo_sigma = 5
    smooth_defo_days: int = 2
    # num_bad_days = 5
    # bad_day_mult = 8
    ref: Tuple[int, int] = (5, 5)

    def make_sar_dates(self):
        """ """
        sar_date_list = []
        for idx in range(self.num_days):
            sar_date_list.append(
                self.start_date + timedelta(days=idx * self.repeat_interval_days)
            )
        return sar_date_list

    def make_sar_stack(
        self,
        seed=None,
    ):
        # Create the 3D stack of turbulence
        sar_stack = self.psd_stack.simulate(
            num_days=self.num_days,
            shape=self.shape,
            randomize=self.randomize,
            seed=seed,
        )
        if self.to_cm:
            sar_stack *= 100

        self.sar_stack = sar_stack
        sar_date_list = self.make_sar_dates()
        self.sar_date_list = sar_date_list
        return sar_stack, sar_date_list

    def make_defo_stack(self, defo_shape="gaussian", **kwargs):
        """Create the time series of deformation to add to each SAR date

        Parameters
        ----------
        defo_shape : str
            Name of a function from `troposim.synthetic`. Defaults to "gaussian".

        Returns
        -------
        defo_stack : np.ndarray (3D)

        """
        from .deformation import synthetic

        # out = np.zeros_like(self.sar_stack)
        # Get the final cumulative deformation
        try:
            defo_func = getattr(synthetic, defo_shape)
        except AttributeError:
            raise ValueError(f"{defo_shape} is not a valid deformation shape")

        # Get shape of deformation in final form (normalized to 1 max)
        final_defo = defo_func(shape=self.shape, **kwargs).reshape((1, *self.shape))
        final_defo /= np.max(final_defo)
        # Broadcast this shape with piecewise linear evolution
        time_evolution = self._make_time_evolution(
            self.defo_rates, smooth=self.smooth_defo_days
        )
        self.defo_stack = final_defo * time_evolution

        return self.defo_stack

    def make_igram_stack(
        self,
        save_ext=None,
        independent=True,
        max_date=None,
        max_date_idx=None,
        max_temporal_baseline=5000,
        sar_stack=None,
        **sar_stack_kwargs,
    ):
        """Create the interferogram stack from the SAR stack.

        Parameters
        ----------
        save_ext : str
            Extension to save the igram stack to. Defaults to None.
        independent : bool
            Whether to select independent or redundant dates. Defaults to True.
        max_date : str
            The maximum date to use in the stack. Defaults to None.
        max_date_idx : int
            The maximum date index to use in the stack. Defaults to None.
        max_temporal_baseline : int
            The maximum temporal baseline (in days) to use in the stack. Defaults to 5000.
        sar_stack : np.ndarray
            The SAR stack to use. Defaults to None.
        sar_stack_kwargs : dict
            Keyword arguments to pass to `make_sar_stack` if `sar_stack` is None.

        Returns
        -------
        igram_stack : np.ndarray (3D)
            The interferogram stack.
        """
        if sar_stack is None:
            if self.sar_stack is None:
                logger.debug("Creating sar_stack")
                self.make_sar_stack(**sar_stack_kwargs)
            sar_stack = self.sar_stack

        if max_date is not None:
            max_date_idx = self.sar_date_list.index(max_date)
        if max_date_idx is None:
            max_date_idx = None

        if independent:
            igram_date_list = self._select_independent(
                self.sar_date_list[:max_date_idx]
            )
        else:
            igram_date_list = self._select_redundant(self.sar_date_list[:max_date_idx])
        # igram_fname_list = []  # list of strings

        igram_stack = []  # list of arrays
        temporal_baselines = []
        selected_dates = []
        for early_date, late_date in igram_date_list:
            if (late_date - early_date).days > max_temporal_baseline:
                continue
            # Using simulated SAR dates to form igrams:
            early_idx = self.sar_date_list.index(early_date)
            late_idx = self.sar_date_list.index(late_date)
            early, late = self.sar_stack[early_idx], self.sar_stack[late_idx]

            selected_dates.append((early_date, late_date))
            # igram = np.abs(late) - np.abs(early) . # Why abs val??
            igram = late - early
            igram_stack.append(igram)
            temporal_baselines.append((late_date - early_date).days)
            if save_ext:
                fname = "{}_{}{}".format(
                    early_date.strftime("%Y%m%d"),
                    late_date.strftime("%Y%m%d"),
                    save_ext,
                )
                igram.tofile(fname)
                # igram_fname_list.append(fname)

        igram_stack = np.array(igram_stack)
        self.igram_date_list = selected_dates
        self.temporal_baselines = np.array(temporal_baselines)
        return igram_stack

    def subtract_reference(self, igram_stack, ref=None):
        """Subtract a reference from the interferogram stack"""
        if ref is None:
            ref = self.ref
        # Add nones to keep 3D
        ref_points = igram_stack[:, ref[0], ref[1]]
        if ref_points.ndim == 1:
            ref_points = ref_points[:, None, None]
        else:
            ref_points = ref_points.mean(axis=(1, 2), keepdims=True)
        return igram_stack - ref_points

    def _make_time_evolution(self, rates, smooth):
        """Create a 1d piecewise linear deformation time evolution
        Reshapes into (1, 1, N) array to broadcast to defo stack
        """
        if np.isscalar(rates):
            rates = [rates]
        rate_idxs = np.linspace(
            0, len(rates), len(self.sar_date_list), endpoint=False
        ).astype(int)
        rates_intp = np.array(rates)[rate_idxs]
        diffs = [d.days for d in np.diff(self.sar_date_list)]

        time_evolution = [0]
        for days, rate in zip(diffs, rates_intp[1:]):
            time_evolution.append(time_evolution[-1] + days * (rate / 365.25))
        time_evolution = np.array(time_evolution)

        if smooth:
            time_evolution = np.convolve(
                time_evolution, np.ones(smooth) / smooth, mode="same"
            )
        return time_evolution.reshape((-1, 1, 1))

    def _select_independent(self, sar_date_list, mid_date=None, num_ifg=None):
        """Choose a list of `num_ifg` independent interferograms spanning `mid_date`

        Parameters
        ----------
        sar_date_list : list[datetime]

        mid_date : datetime, optional
             (Default value = None)
        num_ifg : int, optional
             (Default value = None)

        Returns
        -------
        stack_ifgs : list[(datetime, datetime)]
        """

        if mid_date is not None:
            insert_idx = np.searchsorted(sar_date_list, mid_date)
        else:
            insert_idx = len(sar_date_list) // 2
        num_ifg = num_ifg or len(sar_date_list) - insert_idx

        # Since `mid_date` will fit in the sorted array at `insert_idx`, then
        # sar_date_list[insert_idx] is the first date AFTER the event
        start_idx = np.clip(insert_idx - num_ifg, 0, None)
        end_idx = insert_idx + num_ifg
        date_subset = sar_date_list[start_idx:end_idx]

        stack_ifgs = list(zip(date_subset[:num_ifg], date_subset[num_ifg:]))
        return stack_ifgs

    # TODO: any reason not to do this like Yijue's thesis?
    def _select_redundant(self, sar_date_list):
        """Form all possible interferogram pairs from the date list"""
        return [tuple(pair) for pair in combinations(sar_date_list, 2)]


def create_igrams(num_days=50):
    igm = IgramMaker(num_days=num_days)
    _, sar_date_list = igm.make_sar_stack(seed=None)
    igram_stack, igram_date_list = igm.make_igram_stack()
    return sar_date_list, igram_date_list, igram_stack


def stack_over_time(
    sar_date_list,
    igram_date_list,
    igram_stack,
    start_idx=5,
    skip=2,
    ntotal=9,
    verbose=False,
):
    """Create a series of stacking solutions with progressively longer time window

    Can be plotted with plotting.plot_stack_over_time

    Parameters
    ----------
    sar_date_list : list[datetime]
        list of simulated SAR dates
    igram_date_list : list[tuple(datetime, datetime)]
        list of simulated interferogram dates
    igram_stack : np.ndarray
        simulated interferogram stack
    start_idx : int, optional, Default value = 5
        Index of first date to use in the stack
    skip : int, optional, Default value = 2
        Number of dates to skip between stacks
    ntotal : int, optional
         (Default value = 9)
    verbose : bool, optional
         (Default value = False)

    Returns
    -------
    defos : list[np.ndarray]
    last_dates : list[datetime]
    num_ifgs : list[int]
    num_sar : list[int]
    """

    defos, last_dates, num_ifgs, num_sar = [], [], [], []

    for idx in range(ntotal):
        last_date = sar_date_list[skip * idx + start_idx]
        valid_sar, valid_igrams, valid_idxs = utils.find_valid(
            sar_date_list, igram_date_list, max_date=last_date, verbose=verbose
        )
        phase_sum = np.sum(igram_stack[valid_idxs], axis=0)  # units: cm
        timediffs = [utils.temporal_baseline(ig) for ig in valid_igrams]  # units: days
        avg_velo = phase_sum / np.sum(timediffs)  # cm / day
        # vv = P2MM * avg_velo  # MM / year
        cum_defo = avg_velo * np.max(timediffs)  # cm, cumulative over current interval

        last_dates.append(last_date)
        num_ifgs.append(len(valid_igrams))
        num_sar.append(len(valid_sar))
        defos.append(cum_defo)

    defos = np.stack(defos, axis=0)
    return defos, last_dates, np.array(num_ifgs), np.array(num_sar)



"""Generates surface deformation from fault slip according to Okada's 
rectangular plane model.

Requres okada_wrapper:
https://github.com/tbenthompson/okada_wrapper/
"""
import numpy as np
from scipy.ndimage import rotate

try:
    from okada_wrapper import dc3dwrapper
except ImportError:
    print("okada_wrapper not found. See for instructions:")
    print(" https://github.com/tbenthompson/okada_wrapper/")
    print("Or, pip install okada_wrapper")


def displacement(
    source_depth,
    dip_angle,
    strike_width,
    dip_width,
    alpha=None,
    mu=None,
    lam=None,
    poisson_ratio=0.25,
    strike_angle=0.0,
    rake_angle=0.0,
    slip_magnitude=0.0,
    strike_slip=0.0,
    dip_slip=0.0,
    obs_depth=0.0,
    xy_extent=None,
    shape=None,
    n=200,
):
    """source_depth (float): the depth of the fault origin
    dip_angle (float): the dip-angle, in degrees, of the rectangular dislocation surface
    strike_width (float):  the along-strike width of the surface
    dip_width (float):  the along-dip range of the surface
    alpha = (lambda + mu) / (lambda + 2 * mu)
    dip_slip (float): the dip-slip, in meters, of the rectangular dislocation surface
    strike_slip (float): the strike-slip, in meters, of the rectangular dislocation surface
    rake_angle (float): alternnative to strike_slip and dip_slip, the rake angle of the fault.
        See here for conversions of rake angle to strike_slip and dip_slip:
        https://croninprojects.org/Vince/Course/IntroStructGeol/Pitch.pdf
    slip_magnitude (float): the magnitude of the fault slip, in meters. For use with rake_angle
    obs_depth (float): the depth of the observation point. Defaults to 0.0, the surface
    n (int): the number of points per direction to sample the surface. Defaults to 200.

    Parameters
    ----------
    source_depth :
        
    dip_angle :
        
    strike_width :
        
    dip_width :
        
    alpha :
        (Default value = None)
    mu :
        (Default value = None)
    lam :
        (Default value = None)
    poisson_ratio :
        (Default value = 0.25)
    strike_angle :
        (Default value = 0.0)
    rake_angle :
        (Default value = 0.0)
    slip_magnitude :
        (Default value = 0.0)
    strike_slip :
        (Default value = 0.0)
    dip_slip :
        (Default value = 0.0)
    obs_depth :
        (Default value = 0.0)
    xy_extent :
        (Default value = None)
    shape :
        (Default value = None)
    n :
        (Default value = 200)

    Returns
    -------
    u_out : ndarray
        the displacement field
        shape: (3, ny, nx)
        Each pixel is [ux, uy, uz] 
    xx : ndarray
        the x-coordinates of the displacement field
    yy : ndarray
        the y-coordinates of the displacement field
    
    Examples
    --------
    >>> from okada_wrapper import dc3dwrapper
    >>> success, u, grad_u = dc3dwrapper(0.6, [1.0, 1.0, -1.0],\
                                3.0, 90, [-0.7, 0.7], [-0.7, 0.7],\
                                [1.0, 0.0, 0.0])
    """
    if alpha is None:
        if lam is None:
            if mu is None:
                raise ValueError("Must specify alpha or lam or mu")
            lam = 2 * mu * poisson_ratio / (1 - 2 * poisson_ratio)
        alpha = (mu + lam) / (mu + 2 * lam)

    strike_range = [-strike_width / 2, strike_width / 2]
    dip_range = [-dip_width / 2, dip_width / 2]

    if rake_angle is not None:
        strike_slip, dip_slip = rake_to_slips(rake_angle, slip_magnitude)
    # 3rd is "opening" of fault
    slip_vector = np.array([strike_slip, dip_slip, 0.0])

    if xy_extent is None:
        xy_extent = [-strike_width, -strike_width, strike_width, strike_width]
    left, bot, right, top = xy_extent
    if shape is not None:
        ny, nx = shape
    else:
        ny = nx = n
    xx, yy = np.linspace(left, right, ny), np.linspace(bot, top, nx)

    u_out = np.zeros((3, n, n))
    for j, x in enumerate(xx.ravel()):
        for i, y in enumerate(yy.ravel()):
            success, u, grad_u = dc3dwrapper(
                alpha,
                [x, y, obs_depth],
                source_depth,
                dip_angle,
                strike_range,
                dip_range,
                slip_vector,
            )
            if success != 0:
                raise ValueError(f"dc3dwrapper failed: {success}")
            u_out[:, i, j] = u

    if strike_angle:
        # Also, idk if it's the right north-angle convention.
        # rotates clockwise when plotted with pcolormesh X/Y
        # TODO: make this -90 by default? So it's north up? (assuming y is north)
        u_out = rotate(
            u_out, strike_angle, axes=(-1, -2), reshape=False, mode="nearest"
        )
    return u_out, xx, yy


def project_to_los(u, los):
    """Projects a displacement field to a line of sight.

    Parameters
    ----------
    u : ndarray
        the displacement field from `displacement`
        shape: (3, ny, nx): [ux, uy, uz]
    los : ndarray
        the line of sight unit vector at the observation point
        Can be a single vector, or an array the same size as u.
        shape: (3,): [losx, losy, losz]
        Can be the "ENU" vector of coefficients.

    Returns
    -------
    u_out : 2D ndarray
        the projected displacement field
    """
    los = np.array(los)
    # assert this?
    # los = los / np.linalg.norm(los)
    if los.shape == (3,):
        los = los.reshape(3, 1, 1)
    elif los.shape != u.shape:
        raise ValueError("los must be a 3-vector or the same size as u")
    u_los = np.sum(u * los, axis=0)
    return u_los


def rake_to_slips(rake_angle, slip_magnitude):
    """Convert a rake angle and magnitude to strike-slip and dip-slip components.

    Parameters
    ----------
    rake_angle : float
        the rake angle of the fault, in degrees
        
    slip_magnitude : float
        the magnitude of the fault slip, in meters. For use with rake_angle

    Returns
    -------
    strike_slip : float
        the strike-slip, in meters, of the rectangular dislocation surface
    dip_slip : float
        the dip-slip, in meters, of the rectangular dislocation surface

    
    Examples
    --------
    >>> tuple(np.round(rake_to_slips(-90, 0.1), 4)) # purely dip slip on normal fault
    (0.0, -0.1)
    >>> tuple(np.round(rake_to_slips(0, 0.1), 4)) # pure left lateral strike slip
    (0.1, 0.0)
    >>> tuple(np.round(rake_to_slips(90, 0.1), 4)) # reverse fault slip
    (0.0, 0.1)
    """
    strike_slip = slip_magnitude * np.cos(np.deg2rad(rake_angle))
    dip_slip = slip_magnitude * np.sin(np.deg2rad(rake_angle))
    return strike_slip, dip_slip


def random_displacement(
    source_depth_range=(0.3e3, 5e3),
    dip_angle_range=(0, 90),
    strike_angle_range=(0, 360),
    strike_width_range=(5e3, 20e3),
    dip_width_range=(300, 1000),
    rake_angle_range=(-89, -91),  # dip slip
    slip_magnitude_range=(0.05, 0.5),
    alpha=0.66,
    resolution=180,
    shape=(200, 200),
    seed=None,
    **kwargs,
):
    """

    Parameters
    ----------
    source_depth_range : tuple[float, float]
        (Default value = (0.3e3,5e3)
    dip_angle_range : tuple[float, float]
        (Default value = (0, 90)
    strike_angle_range : tuple[float, float]
        (Default value = (0, 360)
    strike_width_range : tuple[float, float]
        (Default value = (5e3, 20e3)
    dip_width_range : tuple[float, float]
        (Default value = (300, 1000)
    rake_angle_range : tuple[float, float]
        (Default value = (-89, -91)
    slip_magnitude_range : tuple[float, float]
        (Default value = (0.05, 0.5)
    alpha : float
        (Default value = 0.66)
    resolution : float
        (Default value = 180)
    shape : tuple[int, int]
        (Default value = (200, 200)
        
    seed : int
        (Default value = None)

    Returns
    -------
    ndarray
    """
    # see https://stackoverflow.com/a/49849045/4174466 for why RandomState
    rng = np.random.default_rng(seed=seed)
    kwargs["source_depth"] = rng.uniform(*source_depth_range)
    kwargs["dip_angle"] = rng.uniform(*dip_angle_range)
    kwargs["strike_angle"] = rng.uniform(*strike_angle_range)
    kwargs["strike_width"] = rng.uniform(*strike_width_range)
    kwargs["dip_width"] = rng.uniform(*dip_width_range)
    kwargs["rake_angle"] = rng.uniform(*rake_angle_range)
    kwargs["slip_magnitude"] = rng.uniform(*slip_magnitude_range)
    # print(kwargs)

    kwargs["alpha"] = alpha
    ny, nx = shape
    xy_extent = [
        -nx / 2 * resolution,
        -ny / 2 * resolution,
        nx / 2 * resolution,
        ny / 2 * resolution,
    ]
    kwargs["shape"] = shape
    kwargs["xy_extent"] = xy_extent
    return displacement(**kwargs)


def test_okada(
    dip_slip=0.1,
    strike_slip=0.0,
    dip_angle=45.0,
    rake_angle=None,
    slip_magnitude=0.0,
    strike_angle=0.0,
):
    """

    Parameters
    ----------
    dip_slip :
        (Default value = 0.1)
    strike_slip :
        (Default value = 0.0)
    dip_angle :
        (Default value = 45.0)
    rake_angle :
        (Default value = None)
    slip_magnitude :
        (Default value = 0.0)
    strike_angle :
        (Default value = 0.0)

    Returns
    -------
    u : ndarray
        the displacement field from `displacement`
    xx : ndarray
        the x-coordinates of the displacement field
    yy : ndarray
        the y-coordinates of the displacement field
    """
    import matplotlib.pyplot as plt

    # Mentone Eq:
    # strike: 082°, dip: 37°, and rake: −109°. source_depth=5km, 14km along strike
    # Pecos modeling: dip: 57, rake: -90 (all dip-slip), source_depth: .9 km,
    #  15 km along strike, dip_width: .6 km
    # TODO: conversion from rake angle to dip/strike slip vector

    mu = lam = 10e9  # Pa, same as Young's = 25e9 and poisson's = 0.25
    alpha = (lam + mu) / (lam + 2 * mu)
    # In [103]: E, nu = 25e9, 0.25
    # In [106]: sp.solve([mu * (3*lam + 2*mu) / (lam + mu) - E, lam/2/(lam+mu) - nu])
    # Out[106]: [{lam: 10000000000.0000, mu: 10000000000.0000}]
    depth = 900  # meters
    strike_width = 14e3  # meters
    dip_width = 600  # meters

    # dip_slip = 0.1  # meters
    # strike_slip = 0.0  # meters

    u, xx, yy = displacement(
        depth,
        dip_angle,
        strike_width,
        dip_width,
        alpha=alpha,
        dip_slip=dip_slip,
        strike_slip=strike_slip,
        rake_angle=rake_angle,
        slip_magnitude=slip_magnitude,
        strike_angle=strike_angle,
    )

    uz = u[-1]
    vm = np.abs(uz).max()
    plt.figure()
    plt.pcolormesh(xx, yy, u[-1], vmax=vm, vmin=-vm, cmap="RdBu")
    # plt.imshow(u[-1], vmax=vm, vmin=-vm, cmap="RdBu")
    plt.colorbar()
    return u, xx, yy



import numpy as np
from scipy import integrate, special


def reservoir_subsidence(D, R, deltaH, nu=0.25, r=None, as_grid=False):
    """Compute the vertical subsidence according to cylindrical reservoir (Geertsma, 1973).

    Parameters
    ----------
    D : float
        Depth of the reservoir (m).
    R : float
        Radius of the reservoir (m).
    deltaH : float
        Change in reservoir height (m).
    nu : float
        Poisson's ratio. Defaults to 0.25.
    r : ndarray[float] optional
        Radius values at which to evaluate subsidence.
        Defaults to 100 points spaced equally between 0 and R.
    as_grid : bool
        Return the subsidence as a square grid. Defaults to False.
        If False, returns two 1d arrays

    Returns
    -------

    
    """
    # We don't need biot's coefficient if we input the DeltaH.
    # This is related to the pressure drop through DeltaH = H * c_m * DeltaP
    # Notes on other way's to get biot coefficient
    # G_fr = E / (2 * (1 + nu))  # Frame Shear Modulus
    # K_fr = (2 / 3) * G_fr * (1 + nu) / (1 - 2 * nu)  # Frame Bulk Modulus
    # Cm = (1 + nu) * (1 - 2 * nu) / (E * (1 - nu))  # Uniaxial Compressibility
    # alpha = 1 - K_fr / K_mineral  # Biot Coefficient

    # Source: https://github.com/Abakumov/MLIB (though this is wrong)
    def f_A(a, rho):
        """

        Parameters
        ----------
        a :
            
        rho :
            

        Returns
        -------

        
        """
        return np.exp(-D * a) * special.j1(a * R) * special.j0(a * rho)

    if r is None:
        r = np.linspace(0, 2 * R, 100)

    if np.isscalar(r):
        r = np.array([r])

    Uz = np.zeros(r.shape)

    # for i in range(len(rho)):
    # for idx, cur_rho in enumerate(rho):
    for idx, cur_r in enumerate(r):
        int_val = integrate.quad(lambda a: f_A(a, cur_r), 0, np.inf)[0]
        Uz[idx] = 2 * (1 - nu) * deltaH * R * int_val

    if as_grid:
        return _make_square_grid(Uz, r)
    else:
        return Uz, r


def _make_square_grid(Uz, r):
    """

    Parameters
    ----------
    Uz :
        
    r :
        

    Returns
    -------

    
    """
    # Uz, r = reservoir_subsidence(D, R, deltaH, nu=nu, r=r)
    n = len(r)

    # Make the corners of the square equal to R
    xvals = np.linspace(-1 / np.sqrt(2), 1 / np.sqrt(2), n) * max(r)
    XX, YY = np.meshgrid(xvals, xvals)
    RR = np.sqrt(XX ** 2 + YY ** 2)

    r_idxs = find_closest_idxs(r, RR)
    Uz_square = Uz[r_idxs]
    return Uz_square, RR


def plot(D, R, deltaH, nu=0.25, r=None):
    """

    Parameters
    ----------
    D :
        
    R :
        
    deltaH :
        
    nu :
        (Default value = 0.25)
    r :
        (Default value = None)

    Returns
    -------

    
    """
    import matplotlib.pyplot as plt

    Uz_square, RR = reservoir_subsidence(D, R, deltaH, nu=nu, r=r, as_grid=True)
    xmin, xmax = np.array([RR.min(), RR.max()]) / np.sqrt(2)
    fig, ax = plt.subplots()
    axim = ax.imshow(-Uz_square, extent=[xmin, xmax, xmin, xmax], origin="lower")
    ax.set_xlabel("x (m)")
    ax.set_ylabel("y (m)")
    ax.set_title("Vertical subsidence (m)")
    fig.colorbar(axim)
    return Uz_square, RR


def find_closest_idxs(known_array, test_array):
    """

    Parameters
    ----------
    known_array :
        
    test_array :
        

    Returns
    -------

    
    """
    index_sorted = np.argsort(known_array)
    known_array_sorted = known_array[index_sorted]
    known_array_middles = (
        known_array_sorted[1:] - np.diff(known_array_sorted.astype("f")) / 2
    )
    idx1 = np.searchsorted(known_array_middles, test_array)
    indices = index_sorted[idx1]
    return indices



"""Module for simple synthetic deformations, not following any real physical model"""

import random
from collections.abc import Sequence

import numpy as np


def gaussian(
    shape: tuple[int, int],
    sigma: float | tuple[float, float],
    row: int | None = None,
    col: int | None = None,
    normalize: bool = False,
    amp: float | None = None,
    noise_sigma: float = 0.0,
) -> np.ndarray:
    """Create a 2D Gaussian of given shape and width.

    Parameters
    ----------
    shape : tuple[int, int]
        (rows, cols) of the output array
    sigma : float or tuple[float, float]
        Standard deviation of the Gaussian.
        If one float provided, makes an isotropic Gaussian.
        Otherwise, uses [sigma_row, sigma_col] to make elongated Gaussian.
    row : int, optional
        Center row of the Gaussian. Defaults to the middle of the array.
    col : int, optional
        Center column of the Gaussian. Defaults to the middle of the array.
    normalize : bool, optional
        Normalize the amplitude peak to 1. Defaults to False.
    amp : float, optional
        Peak height of Gaussian. If None, peak will be 1/(2*pi*sigma^2).
    noise_sigma : float, optional
        Std. dev of random Gaussian noise added to image. Defaults to 0.0.

    Returns
    -------
    np.ndarray
        2D array containing the Gaussian

    """
    rows, cols = shape
    if not isinstance(sigma, tuple):
        sigma = (sigma, sigma)
    sy, sx = sigma

    # Set default center if not provided
    if row is None:
        row = rows // 2
    if col is None:
        col = cols // 2

    # Create coordinate grids
    y, x = np.ogrid[:rows, :cols]

    # Calculate the 2D Gaussian
    g = np.exp(
        -((x - col) ** 2.0 / (2.0 * sx**2.0) + (y - row) ** 2.0 / (2.0 * sy**2.0))
    )
    normed = _normalize_gaussian(g, normalize=normalize, amp=amp)
    if noise_sigma > 0:
        normed += noise_sigma * np.random.standard_normal(shape)
    return normed


def _rotation_matrix(theta):
    """CCW rotation matrix by `theta` degrees"""
    theta_rad = np.deg2rad(theta)
    return np.array(
        [
            [np.cos(theta_rad), np.sin(theta_rad)],
            [-np.sin(theta_rad), np.cos(theta_rad)],
        ]
    )


def _normalize_gaussian(out, normalize=False, amp=None):
    """Normalize either to 1 max, or to `amp` max"""
    if normalize or amp is not None:
        out /= out.max()
    if amp is not None:
        out *= amp
    return out


def _calc_ab(sigma, ecc):
    """Calculate semi-major/semi-minor axis length from `sigma` and `ecc`entricity"""
    a = np.sqrt(sigma**2 / (1 - ecc))
    b = a * (1 - ecc)
    return a, b


def _xy_grid(shape, xmin=None, xmax=None, ymin=None, ymax=None):
    """Make an xy grid centered at 0,0 in the middle"""
    if xmin is None or xmax is None:
        xmin, xmax = (1, shape[1])
    if ymin is None or ymax is None:
        ymin, ymax = (1, shape[0])
    xx = np.linspace(xmin, xmax, shape[1])
    yy = np.linspace(ymin, ymax, shape[0])
    return np.meshgrid(xx, yy)


def gaussian_ellipse(
    shape: tuple[int, int],
    a: float | None = None,
    b: float | None = None,
    sigma: float | None = None,
    ecc: float | None = None,
    row: int | None = None,
    col: int | None = None,
    theta: float = 0,
    normalize: bool = False,
    amp: float | None = None,
    noise_sigma: float = 0,
) -> np.ndarray:
    """Make an ellipse using multivariate gaussian.

    Parameters
    ----------
    shape : tuple[int, int]
        size of grid
    a : float | None
        semi major axis length
    b : float | None
        semi minor axis length
    sigma : float | None
        std dev of gaussian, if it were circular
    ecc : float | None
        from 0 to 1, alternative to (a, b) specification is (sigma, ecc)
        ecc = 1 - (b/a), and area = pi*sigma**2 = pi*a*b
    row : int | None
        row of center
    col : int | None
        col of center
    theta : float
        degrees of rotation (CCW)
    normalize : bool
        if true, set max value to 1
    amp : float | None
        value of peak of gaussian
    noise_sigma : float
        optional, adds gaussian noise to blob (Default value = 0)

    """
    from scipy.stats import multivariate_normal

    if row is None or col is None:
        nrows, ncols = shape
        row, col = nrows // 2, ncols // 2

    if sigma is not None and ecc is not None:
        a, b = _calc_ab(sigma, ecc)
    if a is None or b is None:
        raise ValueError("Need a,b or sigma,ecc")

    R = _rotation_matrix(theta)
    # To rotate, we do R @ P @ R.T to rotate eigenvalues of P = S L S^-1
    cov = np.dot(R, np.array([[b**2, 0], [0, a**2]]))
    cov = np.dot(cov, R.T)
    var = multivariate_normal(mean=[col, row], cov=cov)

    xx, yy = _xy_grid(shape)
    xy = np.vstack((xx.flatten(), yy.flatten())).T
    out = var.pdf(xy).reshape(shape)
    normed = _normalize_gaussian(out, normalize=normalize, amp=amp)
    if noise_sigma > 0:
        normed += noise_sigma * np.random.standard_normal(shape)
    return normed


def valley(shape, rotate=0):
    """Make a valley in image center (curvature only in 1 direction)"""
    from skimage import transform

    rows, cols = shape
    out = np.dot(np.ones((rows, 1)), (np.linspace(-1, 1, cols) ** 2).reshape((1, cols)))
    if rotate > 0:
        out = transform.rotate(out, None, mode="edge")
    return out


def bowl(shape):
    """Simple quadratic bowl"""
    xx, yy = _xy_grid(shape)
    z = xx**2 + yy**2
    return z / np.max(z)


def quadratic(shape, coeffs):
    """2D quadratic function"""
    nrows, ncols = shape
    row_block, col_block = np.mgrid[0:nrows, 0:ncols]
    yy, xx = row_block.flatten(), col_block.flatten()
    idx_matrix = np.c_[np.ones(xx.shape), xx, yy, xx * yy, xx**2, yy**2]
    return np.dot(idx_matrix, coeffs).reshape(shape)


def stack(N=501, max_amp=3, plot=False, cmap="jet"):
    """Simpler composite of 3 areas of blob of different sizes

    Parameters
    ----------
    N :
        (Default value = 501)
    max_amp :
        (Default value = 3)
    plot :
        (Default value = False)
    cmap :
        (Default value = "jet")

    Returns
    -------


    """
    shape = (N, N)
    b1 = gaussian(shape, 60, N // 3, 2 * N // 3)

    b2 = gaussian(shape, 20, N // 3, N // 3)

    # Group of 4 close ones
    b4 = gaussian(shape, 29, 345, 345)
    b4 += gaussian(shape, 29, 73 + 345, 345)
    b4 += gaussian(shape, 29, 345, 73 + 345)
    b4 += gaussian(shape, 29, 73 + 345, 73 + 345)
    out = b1 - 0.65 * b2 + 0.84 * b4
    out *= max_amp / np.max(out)

    if plot:
        import matplotlib.pyplot as plt

        plt.figure()
        plt.imshow(out, cmap=cmap)
        plt.colorbar()
    return out


def check_overlap(
    new_blob: tuple[int, int, float],
    current_blobs: Sequence[tuple[int, int, float]],
    min_distance_factor: float = 3.0,
) -> bool:
    """
    Check if a new Gaussian blob would overlap with existing ones.

    Parameters
    ----------
    new_blob : tuple[int, int]
        Potential new blob (row, col, sigma).
    current_blobs : Sequence[tuple[int, int]]
        Sequence of existing blob centers (row, col, sigma).
    min_distance_factor : float, default=3.0
        Factor multiplied by the sum of sigmas to determine minimum distance.

    Returns
    -------
    bool
        True if the new blob would NOT overlap with existing ones.
    """
    new_row, new_col, new_sigma = new_blob
    for row, col, sigma in current_blobs:
        # Calculate distance between centers
        distance = np.sqrt((row - new_row) ** 2 + (col - new_col) ** 2)
        # Minimum distance to avoid overlap (using 3-sigma rule)
        min_distance = min_distance_factor * (sigma + new_sigma)
        if distance < min_distance:
            return False
    return True


def multiple_gaussians(
    shape: tuple[int, int],
    num_blobs: int,
    sigma_range: tuple[float, float],
    amp_range: tuple[float, float] | None = None,
    avoid_overlap: bool = True,
    min_distance_factor: float = 3.0,
    border_buffer: int | None = None,
    normalize: bool = True,
    noise_sigma: float = 0.0,
    max_attempts: int = 1000,
    seed: int | None = None,
) -> tuple[np.ndarray, list[tuple[int, int, float]]]:
    """
    Create an image with multiple Gaussian blobs.

    Parameters
    ----------
    shape : tuple[int, int]
        Image dimensions (rows, cols).
    num_blobs : int
        Number of Gaussian blobs to create.
    sigma_range : tuple[float, float]
        Range (min, max) for random sigma values.
    amp_range : tuple[float, float], optional
        Range (min, max) for random amplitude values.
        If None, all blobs have amplitude 1.0.
    avoid_overlap : bool, default=True
        If True, ensures blobs don't overlap significantly.
    min_distance_factor : float, default=3.0
        Minimum distance between blobs as a factor of their combined sigma.
        Only used when avoid_overlap is True.
    border_buffer : int, optional
        Buffer space from image edges where blobs cannot be centered.
        If None, uses half of `max(sigma_range)`.
        Default is None.
    normalize : bool, default=True
        Normalize each Gaussian blob's peak to its amplitude.
    noise_sigma : float, default=0.0
        Standard deviation of random Gaussian noise added to the final image.
    max_attempts : int, default=1000
        Maximum attempts to place a non-overlapping blob before giving up.
    seed: int, optional
        Seed for random generation.
        Default is None (no seeding)

    Returns
    -------
    image: np.ndarray
        2D array containing multiple Gaussian blobs.
    blob_locations: list[tuple[int, int, float]]
        The list of (row, col, sigma) blobs placed in the image

    Notes
    -----
    If avoid_overlap is True and the function cannot place all blobs without
    overlap after max_attempts, it will return with fewer blobs than requested.
    """
    image = np.zeros(shape, dtype=float)
    blob_locations: list[tuple[int, int, float]] = []

    rows, cols = shape
    min_sigma, max_sigma = sigma_range

    # Set default amplitude range if not provided
    if amp_range is None:
        amp_range = (1.0, 1.0)

    if border_buffer is None:
        border_buffer = int(max_sigma // 2)
    # Adjusted placement boundaries accounting for border buffer
    min_row, max_row = border_buffer, rows - border_buffer - 1
    min_col, max_col = border_buffer, cols - border_buffer - 1

    blobs_placed = 0

    if seed:
        random.seed(seed)
    for _ in range(num_blobs):
        # Randomly select parameters for this blob
        sigma = random.uniform(min_sigma, max_sigma)
        amp = random.uniform(amp_range[0], amp_range[1])

        if avoid_overlap and blobs_placed > 0:
            # Try to find a non-overlapping position
            for attempt in range(max_attempts):
                row = random.randint(min_row, max_row)
                col = random.randint(min_col, max_col)

                if check_overlap(
                    (row, col, sigma), blob_locations, min_distance_factor
                ):
                    break

                if attempt == max_attempts - 1:
                    # Couldn't place the blob without overlap after max attempts
                    print(
                        f"Warning: Could only place {blobs_placed} blobs without overlap."
                    )
                    break
            else:
                # This executes if the for loop completes normally (not via break)
                continue
        else:
            # No overlap checking, just place randomly
            row = random.randint(min_row, max_row)
            col = random.randint(min_col, max_col)

        # Add the blob to the image
        blob = gaussian(
            shape,
            sigma,
            row=row,
            col=col,
            normalize=normalize,
            amp=amp,
            noise_sigma=0.0,  # We'll add noise to the final image
        )

        image += blob
        blob_locations.append((row, col, sigma))
        blobs_placed += 1

    # Add noise to the final image if requested
    if noise_sigma > 0:
        image += noise_sigma * np.random.standard_normal(shape)

    return image, blob_locations


def create_cylinder(shape, resolution, bowl_size_meters, bowl_amplitude):
    """
    Create a 2D array representing a cylinder/bowl in the middle of the image.

    Parameters:
    - shape: tuple (height, width) of the image in pixels
    - resolution: spatial resolution in meters per pixel
    - bowl_size_meters: diameter of the bowl in meters
    - bowl_amplitude: height/depth of the bowl

    Returns:
    - 2D numpy array with the bowl shape
    """
    height, width = shape
    bowl_size_pixels = bowl_size_meters / resolution

    # Create coordinate grids
    y, x = np.ogrid[:height, :width]

    # Calculate center coordinates
    center_y, center_x = height // 2, width // 2

    # Calculate distance from center for each pixel
    dist_from_center = np.sqrt((y - center_y) ** 2 + (x - center_x) ** 2)

    # Create the bowl shape - zero outside the bowl radius, amplitude inside
    bowl = np.zeros(shape)
    bowl[dist_from_center <= bowl_size_pixels / 2] = bowl_amplitude

    return bowl



"""Simulate phase noise from decorrelation."""
import numpy as np
from tqdm.auto import tqdm


def simulate(
    coherence=None, looks=1, nbins=200, rounding_threshold=0.05, show_progress=False
):
    """Simulate decorrelation noise from a given coherence.

    Parameters
    ----------
    coherence : float, ndarray
        Coherence magnitude of phase noise. (Default value = None)
    looks : int
        Number of looks. (Default value = 1)
    nbins : int, optional
        Granularity of phi bins, used for PDF generation. Defaults to 200.
    rounding_threshold : float, optional
        Threshold for rounding coherence to nearest multiple. Defaults to 0.05.
        Rounding occurs to speed up PDF calculations, since nearby coherence values
        produce nearly identical PDFs.

    Returns
    -------
    ndarray : Decorrelation phase noise. Shape is same as `coherence`

    Raises
    ------
    ValueError
        If `coherence` is not a scalar or array.

    References
    ----------
    Hanssen, 2001, Eq. 4.2.24
    (Derived by Barber, 1993, Lee et al., 1994, Joughin and Winebrenner, 1994)
    """
    # Make sure it's an array of at least size 1 (no scalars)
    coherence = np.atleast_1d(np.array(coherence))
    out = np.zeros(coherence.shape)
    # Get a truncated list, for a max of (1 / rounding_threshold) unique values
    coh_rounded = _round_to(np.atleast_1d(coherence), rounding_threshold)
    if np.any(coh_rounded < 0) or np.any(coh_rounded > 1):
        raise ValueError("Coherence must be between 0 and 1")

    coh_rounded = np.clip(coh_rounded, 0.01, 0.99)
    looks = np.clip(looks, 1, 100)

    phi_bins = np.linspace(-np.pi, np.pi, nbins + 1, endpoint=True)
    unique_cohs, counts = np.unique(coh_rounded, return_counts=True)
    # Get the PDF for each unique coherence
    _, pdfs = phase_pdf(unique_cohs, looks, nbins=nbins, phi0=0.0)
    for pdf, coh, count in tqdm(
        zip(pdfs, unique_cohs, counts),
        total=len(unique_cohs),
        disable=not show_progress,
    ):
        # Simulate a number of noise samples equal to the
        # number of pixels with this coherence
        samps = _sample_noise(phi_bins, pdf, size=count)
        # Insert these at the appropriate locations in the output array
        idxs = coh_rounded == coh
        out[idxs] = samps
    return out


def phase_pdf(coherence, looks, nbins=200, phi0=0.0):
    """Compute the PDF of decorrelation phase noise for a given number of looks.

    Uses Eq. 4.2.24 from Hanssen, 2001, using the expression derived in
    Barber (1993), Lee et al. (1994), and Joughin and Winebrenner (1994)

    Parameters
    ----------
    coherence : float, ndarray
        Coherence magnitude of phase noise.
    looks : int
        Number of looks.
    nbins : int, optional
        Granularity of phi bins, used for PDF generation. Defaults to 200.
    phi0 : float, optional
        Center phase. Defaults to 0.0.

    Returns
    -------


    """
    from numpy import cos, pi
    from scipy import special as sc

    coherence = np.atleast_1d(coherence)
    phi = np.linspace(-pi, pi, nbins)
    # Add extra axis at end to make broadcasting work
    coherence = coherence[..., np.newaxis]
    # Keep adding extra axes to phi to match coherence.shape
    for ndim in [1, 2, 3]:
        if coherence.ndim > ndim:
            phi = phi[np.newaxis, :]

    phi -= phi0
    d = (1 - coherence**2) ** looks  # factor used twice
    num = sc.gamma(looks + 0.5) * d * np.abs(coherence) * cos(phi)
    z = coherence**2 * cos(phi) ** 2  # reused in hypergeometric function
    denom = 2 * np.sqrt(pi) * sc.gamma(looks) * (1 - z) ** (looks + 0.5)
    factor2 = d / 2 / pi * sc.hyp2f1(looks, 1, 0.5, z)
    return phi, num / denom + factor2


def _sample_noise(phi_bins, pdf, size=1):
    """Sample decorrelation phase noise for a given pdf, using `phase_pdf`.

    Length of phi_bins must be 1 greater than `len(pdf)`.

    Parameters
    ----------
    phi_bins : ndarray
        end points for the phase histogram bin values
    pdf : ndarray
        probability density from `phase_pdf`
    size : int or tuple of ints, optional
        output size for number of samples to draw

    Returns
    -------
    ndarray
    """
    from scipy import stats

    # phi, pdf = phase_pdf(coherence, looks, phi0, nbins)
    dist = stats.rv_histogram((pdf.ravel(), phi_bins))
    return dist.rvs(size=size)


def _round_to(x, step):
    """Round `x` to the nearest `step`"""
    return step * np.round(x / step)



import numpy as np


def remove_ramp(z, deramp_order=1, mask=np.ma.nomask, copy=True, dtype=np.float32):
    """Estimates a linear plane through data and subtracts to flatten
    
    Used to remove noise artifacts from unwrapped interferograms

    Parameters
    ----------
    z : ndarray
        2D array, interpreted as heights
    deramp_order : int
        degree of surface estimation
        deramp_order = 1 removes linear ramp, deramp_order = 2 fits quadratic surface
        (Default value = 1)
    mask : ndarray
         (Default value = np.ma.nomask)
    copy : bool
         (Default value = True)
    dtype : np.dtype
         (Default value = np.float32)

    Returns
    -------
    ndarray
        flattened 2D array with estimated surface removed

    """
    if z.ndim > 2:
        if mask.ndim > 2:
            errstr = f"mask and z must have same length: {len(mask)} != {len(z)}"
            assert len(mask) == len(
                z
            ), errstr
        else:
            mask = [mask] * len(z)
        return np.stack(
            [
                remove_ramp(layer, deramp_order, m, copy, dtype)
                for layer, m in zip(z, mask)
            ]
        )
    z_masked = z.copy() if copy else z
    # Make a version of the image with nans in masked places
    z_masked[mask] = np.nan
    # Use this constrained version to find the plane fit
    z_fit = estimate_ramp(z_masked, deramp_order)
    # Then use the non-masked as return value
    return (z - z_fit).astype(dtype)


def estimate_ramp(z, deramp_order):
    """Takes a 2D array an fits a linear plane to the data
    
    Ignores pixels that have nan values

    Parameters
    ----------
    z : ndarray
        2D array, interpreted as heights
    deramp_order : int
        degree of surface estimation
        deramp_order = 1 removes linear ramp, deramp_order = 2 fits quadratic surface

    Returns
    -------
    ndarray
        the estimated coefficients of the surface
        For deramp_order = 1, it will be 3 numbers, a, b, c from
        ax + by + c = z
        For deramp_order = 2, it will be 6:
        f + ax + by + cxy + dx^2 + ey^2

    """
    if deramp_order > 2:
        raise ValueError("Order only implemented for 1 and 2")
    # Note: rows == ys, cols are xs
    yidxs, xidxs = matrix_indices(z.shape, flatten=True)
    # c_ stacks 1D arrays as columns into a 2D array
    zflat = z.flatten()
    good_idxs = ~np.isnan(zflat)
    if deramp_order == 1:
        A = np.c_[np.ones(xidxs.shape), xidxs, yidxs]
        coeffs, _, _, _ = np.linalg.lstsq(A[good_idxs], zflat[good_idxs], rcond=None)
        # coeffs will be a, b, c in the equation z = ax + by + c
        c, a, b = coeffs
        # We want full blocks, as opposed to matrix_index flattened
        y_block, x_block = matrix_indices(z.shape, flatten=False)
        z_fit = a * x_block + b * y_block + c

    elif deramp_order == 2:
        A = np.c_[
            np.ones(xidxs.shape), xidxs, yidxs, xidxs * yidxs, xidxs ** 2, yidxs ** 2
        ]
        # coeffs will be 6 elements for the quadratic
        coeffs, _, _, _ = np.linalg.lstsq(A[good_idxs], zflat[good_idxs], rcond=None)
        yy, xx = matrix_indices(z.shape, flatten=True)
        idx_matrix = np.c_[np.ones(xx.shape), xx, yy, xx * yy, xx ** 2, yy ** 2]
        z_fit = np.dot(idx_matrix, coeffs).reshape(z.shape)

    return z_fit


def matrix_indices(shape, flatten=True):
    """Returns a pair of vectors for all indices of a 2D array
    
    Convenience function to help remembed mgrid syntax
    
    Example:

    Parameters
    ----------
    shape :
        
    flatten :
         (Default value = True)

    Returns
    -------

    >>> a = np.arange(12).reshape((4, 3))
        >>> print(a)
        [[ 0  1  2]
         [ 3  4  5]
         [ 6  7  8]
         [ 9 10 11]]
        >>> rs, cs = matrix_indices(a.shape)
        >>> rs
        array([0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3])
        >>> cs
        array([0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2])
        >>> print(a[rs[1], cs[1]] == a[0, 1])
        True
    """
    nrows, ncols = shape
    row_block, col_block = np.mgrid[0:nrows, 0:ncols]
    if flatten:
        return row_block.flatten(), col_block.flatten()
    else:
        return row_block, col_block



import glob
from concurrent.futures import ProcessPoolExecutor, as_completed

import h5py
import numpy as np
import rasterio as rio

from troposim.deformation import okada

from . import decorrelation, deformation, stratified, turbulence, utils

SENTINEL_WAVELENGTH = 5.5465763  # cm


# TODO: utils.record_params_as_yaml
# TODO: downsample as data augmentation
def generate_stacks(
    demfile: str,
    outfile: str = "simulated_stack.h5",
    dsets: list[str] = ["defo", "stratified", "turbulence", "decorrelation"],
    num_days: int = 9,
    num_defos: int = 120,
    defo_shape: tuple[int, int] = (200, 200),
    add_day1_turbulence: bool = False,
    turbulence_kwargs: dict = {},
    stratified_kwargs: dict = {},
    deformation_kwargs: dict = {},
) -> str:
    """
    Generate simulated InSAR stacks with various atmospheric and deformation components.

    Parameters
    ----------
    demfile : str
        Path to the input DEM file.
    outfile : str, optional
        Path to the output HDF5 file, by default "simulated_stack.h5".
    dsets : List[str], optional
        List of datasets to generate, by default ["defo", "stratified", "turbulence", "decorrelation"].
    num_days : int, optional
        Number of days in the time series, by default 9.
    num_defos : int, optional
        Number of deformation events to simulate, by default 120.
    defo_shape : Tuple[int, int], optional
        Shape of each deformation event, by default (200, 200).
    add_day1_turbulence : bool, optional
        Whether to add turbulence on the first day, by default False.
    turbulence_kwargs : Dict, optional
        Additional kwargs for turbulence simulation, by default {}.
    stratified_kwargs : Dict, optional
        Additional kwargs for stratified delay simulation, by default {}.
    deformation_kwargs : Dict, optional
        Additional kwargs for deformation simulation, by default {}.

    Returns
    -------
    str
        Path to the output HDF5 file.

    Notes
    -----
    This function simulates various components of an InSAR time series, including
    stratified atmospheric delay, turbulent atmospheric delay, decorrelation noise,
    and deformation events. The results are saved in an HDF5 file.
    """
    res_30_degrees = 0.000277777777

    with rio.open(demfile) as src:
        dem = src.read(1).astype(np.float32)
        resolution = round(src.transform[0] / res_30_degrees * 30)

    shape = dem.shape
    stratified_kwargs.update({"K_params": {"shape": (num_days,)}})
    print(stratified_kwargs)
    strat = stratified.simulate(dem, **stratified_kwargs)
    print(
        f"Stratified (min, max, mean): {strat.min(), strat.max(), strat.mean(axis=(1, 2))}"
    )

    turb = 100 * turbulence.simulate(
        shape=(num_days, *shape), resolution=resolution, **turbulence_kwargs
    )
    if add_day1_turbulence:
        turb_day1 = 100 * turbulence.simulate(
            shape=shape, resolution=resolution, **turbulence_kwargs
        )
        turb += turb_day1.reshape((1, *shape))
    print(
        f"Turbulence (min, max, mean): {turb.min(), turb.max(), turb.mean(axis=(1, 2))}"
    )

    looks = 50
    coh = np.clip(0.5 + 100 * turbulence.simulate(shape=shape), 0, 1)
    coh = np.tile(coh, (num_days, 1, 1))
    cohphase = decorrelation.simulate(coh, looks) * PHASE_TO_CM_S1
    print(f"Cohphase (min, max): {cohphase.min(), cohphase.max()}")

    defo_stack = np.zeros((num_days, shape[0], shape[1]), dtype=np.float32)

    with ProcessPoolExecutor(max_workers=20) as executor:
        futures = [
            executor.submit(
                deformation.okada.random_displacement,
                resolution=resolution,
                shape=defo_shape,
                **deformation_kwargs,
            )
            for _ in range(num_defos)
        ]
        for future in as_completed(futures):
            defo, _, _ = future.result()
            defo *= 100
            los = _get_random_los()
            defo = okada.project_to_los(defo, los)

            row = np.random.randint(0, shape[0] - defo_shape[0])
            col = np.random.randint(0, shape[1] - defo_shape[1])
            date_idx = np.random.randint(0, num_days)
            defo_stack[
                date_idx:, row : row + defo_shape[0], col : col + defo_shape[1]
            ] += defo
    print(f"Defo stack (min, max): {defo_stack.min(), defo_stack.max()}")

    attrs = {
        "defo": {
            "shape": defo_shape,
            "num_defos": num_defos,
            "kwargs": str(deformation_kwargs),
        },
        "stratified": {"kwargs": str(stratified_kwargs)},
        "turbulence": {
            "add_day1_turbulence": add_day1_turbulence,
            "kwargs": str(turbulence_kwargs),
        },
        "decorrelation": {"looks": looks},
    }
    print("Saving stacks")
    comp_dict = {"compression": 32001, "compression_opts": (0, 0, 0, 0, 5, 1, 1)}
    stacks = [defo_stack, strat, turb, cohphase]
    with h5py.File(outfile, "w") as hf:
        hf.attrs["resolution"] = resolution
        for data, dset in zip(stacks, dsets):
            hf.create_dataset(
                dset,
                shape=data.shape,
                dtype="float32",
                data=data,
                chunks=True,
                **comp_dict,
            )
            for k, v in attrs[dset].items():
                hf[dset].attrs[k] = v
    return outfile


def _get_random_los():
    """Picks a random LOS, similar to Sentinel-1 range"""
    north = 0.1
    up = -1 * np.random.uniform(0.5, 0.85)
    east = np.sqrt(1 - up**2 - north**2)
    # sometimes ascending pointing east, sometimes descending pointing west
    east *= np.random.choice([-1, 1])
    return [east, north, up]


def data_loader(
    sim_file,
    dem_file,
    chunk_size=(48, 48),
    looks=(1, 1),
    normalize=False,
    load=True,
    no_noise=False,
):
    """Create a generator that yields chunks of data from the noise and defo files.

    Parameters
    ----------
    noise_file : str
        Name of the noise file.
    defo_file : str
        Name of the defo file.
    dem_file : str
        Name of the DEM file.
    chunk_size : tuple
        Spatial size of chuns to yield. Defaults to (48, 48).
    looks :
        (Default value = (1, 1)
    normalize : bool
        Whether to normalize the data to [-1, 1].
        Defaults to True.
    load : bool
        Whether to load the data into memory. Defaults to True.
    no_noise : bool
        Whether to skip loading the noise file. Defaults to False.
    sim_file :


    Yields
    -------
    X : 3D np.ndarray
        data input
    y : np.ndarray
        target. Really 2D, but padded to match the shape of X.
    """
    hf = h5py.File(sim_file, "r")
    full_shape = hf["defo"].shape
    defo_stack = hf["defo"]

    if not no_noise:
        strat_stack = hf["stratified"]
        turb_stack = hf["turbulence"]
        decor_stack = hf["decorrelation"]
    else:
        strat_stack = turb_stack = decor_stack = np.zeros_like(defo_stack)
    if load:
        defo_stack = defo_stack[()]
        strat_stack = strat_stack[()]
        turb_stack = turb_stack[()]
        decor_stack = decor_stack[()]
        if looks != (1, 1):
            defo_stack = utils.take_looks(defo_stack, *looks)
            strat_stack = utils.take_looks(strat_stack, *looks)
            turb_stack = utils.take_looks(turb_stack, *looks)
            decor_stack = utils.take_looks(decor_stack, *looks)
        print(f"Shape: {defo_stack.shape}")

    if normalize:
        c1 = np.max(np.abs(defo_stack))
        c2 = np.max(np.abs(strat_stack))
        c3 = np.max(np.abs(turb_stack))
        c4 = np.max(np.abs(decor_stack))
        c = np.max([c1, c2, c3, c4])
    else:
        c = 1.0

    with rio.open(dem_file) as src:
        dem = src.read(1).astype(np.float32)
        if normalize:
            # Scale DEM to be between 0 and 1
            # dem = (dem - dem.min()) / (dem.max() - dem.min())
            # Scale DEM to be zero mean and unit variance
            dem = (dem - dem.mean()) / dem.std()
            # does it make a difference which normalizing?
        if looks != (1, 1):
            dem = utils.take_looks(dem, *looks)

    blk_slices = utils.block_iterator(
        full_shape[-2:], chunk_size, overlaps=(0, 0), start_offsets=(0, 0)
    )

    for rows, cols in blk_slices:
        # with h5py.File(sim_file) as hf:
        dem_chunk = dem[slice(*rows), slice(*cols)].astype(np.float32)
        defo = defo_stack[:, slice(*rows), slice(*cols)]

        strat = strat_stack[:, slice(*rows), slice(*cols)]
        turb = turb_stack[:, slice(*rows), slice(*cols)]
        decor = decor_stack[:, slice(*rows), slice(*cols)]
        noise = strat + turb + decor
        # with rio.open(dem_file) as src:
        #     window = rio.windows.Window.from_slices(rows, cols)
        #     dem_chunk = src.read(1, window=window)
        #     if normalize:
        #         # Turn the DEM into a kilometer scale
        #         dem_chunk = (dem_chunk / 1000).astype(np.float32)
        if defo.shape[-2:] != chunk_size or noise.shape[-2:] != chunk_size:
            continue

        X = (noise + defo) / c
        # Make the DEM the first layer of the stack
        X = np.concatenate([dem_chunk[np.newaxis], X], axis=0)

        # For the target, we don't want to include the deformation that occurs on the first
        # date (since we'll be sliding this along through time and don't want to keep outputting
        # the same deformation), and we also want to ignore the last date's deformation
        target_defo = defo[-2] - defo[0]
        y = target_defo[np.newaxis, :, :] / c
        yield X, y
    hf.close()


def load_all_data(
    sim_file_glob=None,
    dem_file=None,
    sim_file_list=None,
    chunk_size=(48, 48),
    looks=(1, 1),
    normalize=False,
    no_noise=False,
    load=True,
    min_y=0.7,  # cm, unnormalized
    max_y=5,  # cm, unnormalized
):
    """

    Parameters
    ----------
    sim_file_glob :
        (Default value = None)
    dem_file :
        (Default value = None)
    sim_file_list :
        (Default value = None)
    chunk_size :
        (Default value = (48)
    48) :

    looks : tuple[int, int]
        (Default value = (1, 1)

    normalize :
        (Default value = False)
    no_noise :
        (Default value = False)
    load : bool
        (Default value = True)
    min_y : float
        (Default value = 0.7 cm)
    max_y : float
        (Default value = 5 cm)

    """
    if sim_file_list is None:
        sim_file_list = glob.glob(sim_file_glob)
    X_all = None
    y_all = None
    for f in sim_file_list:
        dl = data_loader(
            f,
            dem_file,
            chunk_size=chunk_size,
            normalize=normalize,
            load=load,
            no_noise=no_noise,
            looks=looks,
        )
        rows = list(dl)
        Xs, ys = zip(*rows)
        Xs, ys = np.array(Xs), np.array(ys)
        # TODO: add a random sample of zeros?
        # zero_idxs = [i for i, y in enumerate(ys) if np.sum(np.abs(y)) == 0]
        # num_zeros_to_keep = int(pct_zeros * len(ys))
        axis = (1, 2, 3) if ys.ndim == 4 else (1, 2)
        y_maxes = np.max(np.abs(ys), axis=axis)
        keep_idxs = np.logical_and(y_maxes > min_y, y_maxes < max_y)
        # keep_idxs = [i for i, y in enumerate(ys) if np.max(np.abs(y)) > min_y]
        Xs = Xs[keep_idxs]
        ys = ys[keep_idxs]

        if X_all is None:
            X_all = Xs
            y_all = ys
        else:
            X_all = np.concatenate((X_all, Xs))
            y_all = np.concatenate((y_all, ys))

    return X_all, y_all


def plot_data(X, y, dem, vm=None, cmap="RdBu"):
    import matplotlib.pyplot as plt

    vm = vm or np.max(np.abs(X))
    # fig, axes = plt.subplots(3, 3, figsize=(8, 8))
    fig = plt.figure(constrained_layout=True, figsize=(9, 6))
    gs0 = fig.add_gridspec(1, 2)

    gs_left = gs0[0].subgridspec(3, 3)
    gs_right = gs0[1].subgridspec(2, 1)
    # Input timesteps
    for a in range(3):
        for b in range(3):
            i = 3 * a + b
            ax = fig.add_subplot(gs_left[a, b])
            ax.imshow(X[i], cmap=cmap, vmin=-vm, vmax=vm)
            ax.set_title(f"{i}")

    # Right side: DEM, and target deformation
    # axes = axes.ravel()
    ax = fig.add_subplot(gs_right[0, 0])
    axim = ax.imshow(dem, cmap="terrain")
    fig.colorbar(axim, ax=ax)
    ax.set_title("DEM")

    ax = fig.add_subplot(gs_right[1, 0])
    axim = ax.imshow(y, cmap=cmap, vmax=vm, vmin=-vm)
    fig.colorbar(axim, ax=ax)
    ax.set_title("Target deformation")
    # axes[i].set_title(f"{y[i]:.2f} cm")


def power(x):
    return np.mean(x**2)


def load_noise(hf):
    return hf["turbulence"][()] + hf["stratified"][()] + hf["decorrelation"][()]


def snr(hf):
    return power(hf["defo"][()]) / power(load_noise(hf))


def peak_ratio(hf, axis=None):
    defo_max = np.max(np.abs(hf["defo"][()]), axis=axis)
    noise_max = np.max(np.abs(load_noise(hf)), axis=axis)
    return defo_max / noise_max


def get_stack_metrics(glob_str, func, *args, **kwargs):
    file_list = glob.glob(glob_str)
    metrics = []
    for f in file_list:
        with h5py.File(f) as hf:
            metrics.append(func(hf, *args, **kwargs))
    return file_list, metrics


def get_snrs(glob_str, func=snr):
    return get_stack_metrics(glob_str, func)


def get_peaks(glob_str, func=peak_ratio):
    return get_stack_metrics(glob_str, func)


class MeanMaxScaler:
    def __init__(self):
        pass

    def fit(self, X):
        self.mean_ = np.average(X)
        # self.scale_ = np.max(np.abs(X))
        self.scale_ = np.max(np.abs(X - self.mean_))  # check with Scott on this
        return self

    def transform(self, X):
        return (X - self.mean_) / self.scale_

    def inverse_transform(self, X):
        return X * self.scale_ + self.mean_


class LogMeanScaler:
    """Can work well to normalize skewed data, I tried it but no huge improvement at least
    on very noisy data.
    """

    def __init__(self):
        pass

    def fit(self, X):
        self.shift_ = (
            0 if np.min(X) >= 0 else -np.min(X)
        ) + 0.0001  # log10(0) is undefined so slight shift
        self.mean_ = np.average(np.log(X + self.shift_))
        return self

    def transform(self, X):
        return np.log(X + self.shift_) - self.mean_

    def inverse_transform(self, X):
        return np.exp(X + self.mean_) - self.shift_


class Scaler:
    """ """

    def __init__(self, scale_percentile=95):
        self.mean_, self.scale_ = None, None
        self.dem_mean_, self.dem_scale_ = None, None
        self.scale_percentile = scale_percentile

    def fit(self, X):
        # Divide into dem/timeseries
        dems, timeseries = self._unstack_dem_timeseries(X)

        if dems is not None:
            # Scale DEM to be between -1 and 1
            self.dem_mean_ = np.mean(dems)
            # self.dem_scale_ = np.max(np.abs(dems - self.dem_mean_))
            self.dem_scale_ = np.std(dems)  # For setting unit variance

        self.mean_ = np.mean(timeseries)
        # self.scale_ = np.max(np.abs(timeseries))
        self.scale_ = np.percentile(np.abs(timeseries), self.scale_percentile)

        return self

    def transform(self, X):
        dems, timeseries = self._unstack_dem_timeseries(X)

        if dems is not None:
            if self.dem_mean_ is None:
                raise ValueError("Must fit DEM before transforming")
            dems_scaled = (dems - self.dem_mean_) / self.dem_scale_
            timeseries_scaled = (timeseries - self.mean_) / self.scale_
            return self._stack_dem_timeseries(dems_scaled, timeseries_scaled)
        else:
            return (timeseries - self.mean_) / self.scale_

    def inverse_transform(self, X):
        dems, timeseries = self._unstack_dem_timeseries(X)

        if dems is not None:
            if self.dem_mean_ is None:
                raise ValueError("Must fit DEM before inverse transforming")
            dems_unscaled = dems * self.dem_scale_ + self.dem_mean_
            ts_unscaled = timeseries * self.scale_ + self.mean_
            return self._stack_dem_timeseries(dems_unscaled, ts_unscaled)
        else:
            return timeseries * self.scale_ + self.mean_

    def _unstack_dem_timeseries(self, X):
        layer0 = X[:, :1]
        other_layers = X[:, 1:]
        # The y targets will be shape (batch, 1, rows, cols)
        # So `other_layers` will be shape (batch, 0, rows, cols) after unstacking.

        if other_layers.size > 0:
            dems = layer0
            timeseries = other_layers
        else:
            dems = None
            timeseries = layer0
        return dems, timeseries

    def _stack_dem_timeseries(self, dems, timeseries):
        return np.concatenate((dems, timeseries), axis=1)

    def asdict(self):
        d = {"mean": self.mean_, "scale": self.scale_}
        if self.dem_mean_ is not None:
            d["dem_mean"] = self.dem_mean_
            d["dem_scale"] = self.dem_scale_
        return d


def plot_model_filters(model, layer_num=-1, cmap="plasma", in_idx=0):
    import matplotlib.pyplot as plt

    weights = model.get_weights()[layer_num]
    n_filters = weights.shape[-1]

    n_rows = int(np.ceil(np.sqrt(n_filters)))
    n_cols = int(np.ceil(n_filters / n_rows))
    fig, axes = plt.subplots(n_rows, n_cols, figsize=(10, 10), squeeze=False)

    for ax, i in zip(axes.flat, range(n_filters)):
        vm = np.max(np.abs(weights[..., in_idx, :]))
        filt = np.squeeze(weights[..., in_idx, i])
        filt = filt[0] if len(filt.shape) == 3 else filt
        ax.imshow(filt, cmap=cmap, vmax=vm, vmin=-vm)
        ax.axis("off")
        ax.set_title("Filter {}".format(i))


def save_predictions_as_tifs(
    arr, ds_file, idxs, outfile_template="{inp}_predicted_{d1}_{d2}.tif"
):
    import xarray as xr
    from apertools import sario

    with xr.open_dataset(ds_file) as ds:
        lats = ds["lat"]
        lons = ds["lon"]
        date_ranges = ds["date_ranges"]

    for idx, img in zip(idxs, arr):
        dates = date_ranges[idx].dt.strftime("%Y%m%d")
        d1, d2 = dates[0].item(), dates[-1].item()
        da = xr.DataArray(img, coords={"lat": lats, "lon": lons}, dims=["lat", "lon"])

        outfile = outfile_template.format(inp=ds_file.replace(".nc", ""), d1=d1, d2=d2)
        sario.save_xr_tif(outfile, da)



try:
    import matplotlib.pyplot as plt
except ImportError:
    print("matplotlib not available, skipping plotting")
import numpy as np

DATE_FMT = "%Y%m%d"
CMAP = "RdBu_r"


def plot_stack(stack, titles=None, ntotal=9, cmap=CMAP, figsize=(8, 8)):
    """Plot a portion of a stack of images in a facet grid

    Parameters
    ----------
    stack : ndarray
        3D array of images
    titles : list of str
        (Default value = None)
    ntotal : int
        (Default value = 9)
    cmap :
        (Default value = CMAP)
    figsize : tuple[int, int]
        (Default value = (8)
    """
    nrow = ncol = int(np.sqrt(ntotal))
    fig, axes = plt.subplots(nrow, ncol, figsize=figsize)
    vmin = np.min(stack)
    vmax = np.max(stack)
    for idx, ax in enumerate(axes.ravel()):
        axim = ax.imshow(stack[idx], cmap=cmap, vmax=vmax, vmin=vmin)
        fig.colorbar(axim, ax=ax)
        if titles is not None:
            ax.set_title(titles[idx])


def plot_psd(
    data,
    freq=None,
    psd1d=None,
    freq0=1e-4,
    resolution=None,
    axes=None,
    outfig=None,
    cmap=CMAP,
    color="k",
    label=None,
    marker=None,
    in_mm=False,
    per_km=True,
    slopes=[],
    vm=None,
):
    """Plot the result of `turbulence.get_psd`

    Parameters
    ----------
    data : ndarray
        Noise image to plot the PSD of
    freq : ndarray, optional
        Frequecny array to use for plot
        if not passed, will be generated from the data
    psd1d : ndarray
        PSD to use for plot
        if not passed, will be generated from the data
    freq0 : float, default value = 1e-4
        reference frequency line to plot, showing p0
    resolution : float, optional
        Resolution of data (if not passing freq + psd1d)
    axes : matplotlib.axes.Axes
    outfig : str
        path to save the figure to
    cmap : str, default value = RdBu_r
        Colormap to use for image plot
    color : str, default value = "k"
        Line color for plot
    label : str, optional
        Label for plot
    marker : str, optional
        Marker for plot
    in_mm : bool, default value = False
        If True, plot in mm instead of cm
    per_km : bool, default value = True
        If True, plot in units of 1/km instead of 1/m
    slopes : list[float], default value = []
        Linear slopes to plot on the PSD for reference
    vm : float, optional
        Max value for image plot


    Returns
    -------
    fig : matplotlib.figure.Figure
        Figure object
    axes : matplotlib.axes.Axes
        Axes object
    """
    if freq is None or psd1d is None:
        from . import turbulence

        if resolution is None:
            raise ValueError("resolution must be specified if freq and psd1d are not")

        _, _, freq, psd1d = turbulence.get_psd(
            data,
            resolution=resolution,
            freq0=freq0,
        )

    if axes is None:
        fig, axes = plt.subplots(nrows=1, ncols=2, figsize=[10, 3])
    else:
        ax = axes[0]
        fig = ax.get_figure()

    if vm is None:
        vm = np.nanmax(np.abs(data * 100))
    axim = axes[0].imshow(data * 100, cmap=cmap, vmin=-vm, vmax=vm)
    axes[0].set_axis_off()
    cbar = fig.colorbar(axim, ax=axes[0])
    cbar.set_label("cm")

    plot_psd1d(
        freq,
        psd1d,
        ax=axes[1],
        freq0=freq0,
        label=label,
        color=color,
        marker=marker,
        per_km=per_km,
        in_mm=in_mm,
        slopes=slopes,
    )

    fig.tight_layout()

    if outfig:
        fig.savefig(outfig, bbox_inches="tight", transparent=True, dpi=300)
        print("save figure to", outfig)
    return fig, axes
    # plt.show()


def plot_psd1d(
    freq,
    psd1d,
    ax=None,
    freq0=None,
    label=None,
    color="k",
    linestyle="-",
    marker=None,
    per_km=True,
    in_mm=False,
    slopes=[],
    **plotkwargs,
):
    if ax is None:
        fig, ax = plt.subplots()

    # denom: go from dividing by (1/m)**2 to (1/km)**2 -> (1000 m/1km)**2
    denom_scale = 1e6 if per_km else 1
    numer_scale = 1e4  # from (meters**2 to cm**2)
    if in_mm:
        # extra (10^2) for cm->mm
        numer_scale *= 100
    freq_plot = freq * (1e3 if per_km else 1)
    ax.loglog(
        freq_plot,
        psd1d * numer_scale / denom_scale,
        label=label,
        color=color,
        linestyle=linestyle,
        marker=marker,
        **plotkwargs,
    )
    if freq0 is not None:
        # vertical line marking reference frequency
        ax.axvline(freq0 * 1e3, linestyle="--", color="k")

    # ax.set_xlabel("Wavenumber [cycle/km]")
    ax.set_xlabel(r"$k$ [cycle/km]")
    ylabel = "PSD "
    ylabel += r"$[cm^2 / (1/m^2)]$"
    if per_km:
        ylabel = ylabel.replace("/m", "/km")
    if in_mm:
        ylabel = ylabel.replace("cm", "mm")
    ax.set_ylabel(ylabel)
    # If passed, plot the slopes to compare to the PSD
    for slope in slopes:
        slope_color = "orange"
        lines = freq ** slope
        lines *= numer_scale / denom_scale * (np.max(psd1d) / lines[0])
        ax.loglog(freq_plot, .1 * lines, color=slope_color, linestyle="--")
        ax.loglog(freq_plot, 30 * lines, color=slope_color, linestyle="--")
    return ax


def plot_stack_over_time(
    avgs,
    sar_dates_used,
    vm=None,
    igram_stack=None,  # Just for colorbar
    cmap="seismic_wide_y",
):
    if vm is None:
        if igram_stack is not None:
            vm = np.nanpercentile(np.nanmax(np.abs(igram_stack)), 80)
        else:
            vm = 2

    ntotal = len(avgs)
    nrow = ncol = int(np.ceil(np.sqrt(ntotal)))
    fig, axes = plt.subplots(nrow, ncol)

    for vv, last_date, ax in zip(avgs, sar_dates_used, axes.ravel()):
        axim = ax.imshow(vv, vmax=vm, vmin=-vm, cmap=cmap)
        ax.set_title(last_date.strftime(DATE_FMT))
        ax.set_axis_off()
        fig.colorbar(axim, ax=ax)

    return fig, axes


def plot_psd1d_azimuth(
    angles,
    psd1d,
    ax=None,
    label=None,
    ylim=(0, 1),
    color="k",
    # linestyle="-",
    marker="o",
    **plotkwargs,
):
    if not ax:
        fig, ax = plt.subplots()

    ax.plot(
        angles,
        psd1d,
        label=label,
        color=color,
        # linestyle=linestyle,
        marker=marker,
        **plotkwargs,
    )
    from . import utils

    a1 = utils.anisotropy_var(psd1d)
    a2 = utils.anisotropy_A(psd1d)
    title = f"A = {a2:.2f}, var = {a1:.2g}"
    ax.set_title(title)

    ax.set_ylim(ylim)
    ax.set_xlabel("Angular bin [degrees]")
    ax.set_ylabel("Average Power")



import numpy as np

DATE_FMT = "%Y%m%d"


def crop_square_data(data, N=None):
    """Grab the largest square portion of size N from input 2D input

    Searches for square with greatest number of non-zero values.

    Parameters
    ----------
    data : 2D ndarray
        input image

    N : int optional
        Size of square input.
        If not provided, uses N = min(data.shape)
        Uses the integral image (summed area table) of data to quickly
        calculate number of zeros in a square subset:
        https://en.wikipedia.org/wiki/Summed-area_table (Default value = None)

    Returns
    -------
    ndarray
        square subset of data, size = (N, N)

    """
    row_slice, col_slice = find_largest_square(data, N=N)
    return data[row_slice, col_slice]


def find_largest_square(data, N=None):
    """Get the row and column slices of the largest square in a 2D array

    Parameters
    ----------
    data : 2D ndarray

    N : int
        size of square to search for (Default value = min(data.shape))

    Returns
    -------
    tuple[slice, slice]

    """
    if N is None:
        N = min(data.shape)
    if N > min(data.shape):
        raise ValueError(
            f"N={N} must be less than the smallest dimension: {min(data.shape)}"
        )

    nonzeros = np.logical_and(data != 0, ~np.isnan(data))
    # Use integral image for quick rectangle area summing of # of True
    S = integral_image(nonzeros)
    row = col = 0
    max_nonzeros = 0
    max_row, max_col = 0, 0
    # Move square subset around to test
    # NOTE: you only ever need to move in one direction
    # (since N == one of the dimensions)
    # so this runs quickly
    while row + N - 1 < data.shape[0]:
        while col + N - 1 < data.shape[1]:
            # See https://en.wikipedia.org/wiki/Summed-area_table
            A = (row, col)
            B = (row + N - 1, col)
            C = (row, col + N - 1)
            D = (row + N - 1, col + N - 1)
            num_nonzeros = S[A] + S[D] - S[B] - S[C]

            if num_nonzeros > max_nonzeros:
                max_nonzeros = num_nonzeros
                max_row, max_col = row, col
            col += 1
        col = 0
        row += 1
    return slice(max_row, max_row + N), slice(max_col, max_col + N)


def integral_image(image):
    r"""Integral image / summed area table.

    The integral image contains the sum of all elements above and to the
    left of it, i.e.:

    .. math::

       S[m, n] = \sum_{i \leq m} \sum_{j \leq n} X[i, j]

    Parameters
    ----------
    image : ndarray
        Input image.

    Returns
    -------
    S : ndarray
        Summed area table, same size as `image`

    Source
    ------
    skimage.transform module

    References
    ----------
    .. [1] F.C. Crow, "Summed-area tables for texture mapping,"
           ACM SIGGRAPH Computer Graphics, vol. 18, 1984, pp. 207-212.
    """
    S = image.copy()
    for i in range(image.ndim):
        S = S.cumsum(axis=i)
    return S


def extrema(x):
    """Get the (min, max) of an array"""
    return np.min(x), np.max(x)


def find_valid(
    sar_date_list,
    ifg_date_list,
    min_date=None,
    max_date=None,
    max_temporal_baseline=900,
    verbose=True,
):
    """Cut down the full list of interferograms and geo dates

    - Cuts date ranges (e.g. piecewise linear solution) with  `min_date` and `max_date`
    - Can ignore whole dates by reading `ignore_geo_file`
    - Prunes long time baseline interferograms with `max_temporal_baseline`

    Parameters
    ----------
    sar_date_list : iterable of datetime

    ifg_date_list : iterable of (datetime, datetime)

    min_date : datetime
         (Default value = None)
    max_date : datetime
         (Default value = None)
    max_temporal_baseline : int
         (Default value = 900)
    verbose : bool
         (Default value = True)

    Returns
    -------
    valid

    """
    if verbose:
        ig1 = len(ifg_date_list)  # For logging purposes, what do we start with

    valid_slcs = sar_date_list
    valid_ifgs = ifg_date_list

    # Remove slcs and ifgs outside of min/max range
    if min_date is not None:
        if verbose:
            print(f"Keeping data after min_date: {min_date}")
        valid_slcs = [g for g in valid_slcs if g > min_date]
        valid_ifgs = [
            ig for ig in valid_ifgs if (ig[0] > min_date and ig[1] > min_date)
        ]

    if max_date is not None:
        if verbose:
            print(f"Keeping data only before max_date: {max_date}")
        valid_slcs = [g for g in valid_slcs if g < max_date]
        valid_ifgs = [
            ig for ig in valid_ifgs if (ig[0] < max_date and ig[1] < max_date)
        ]

    if verbose:
        # This is just for logging purposes:
        too_long_ifgs = [
            ig for ig in valid_ifgs if temporal_baseline(ig) > max_temporal_baseline
        ]
        print(
            f"Ignoring {len(too_long_ifgs)} ifgs with longer baseline "
            f"than {max_temporal_baseline} days"
        )

    # ## Remove long time baseline ifgs ###
    valid_ifgs = [
        ig for ig in valid_ifgs if temporal_baseline(ig) <= max_temporal_baseline
    ]

    if verbose:
        print(f"Ignoring {ig1 - len(valid_ifgs)} ifgs total")

    # Now go back to original full list and see what the indices were
    # used for subselecting from the unw_stack by a pixel
    valid_idxs = np.searchsorted(
        intlist_to_string(ifg_date_list),  # Numpy searchsorted can't use datetimes
        intlist_to_string(valid_ifgs),
    )
    return valid_slcs, valid_ifgs, valid_idxs


def temporal_baseline(ifg):
    """Get the temporal baseline (int days) of an interferogram

    Parameters
    ----------
    ifg : tuple[datetime, datetime]

    Returns
    -------
    T : int

    """
    return (ifg[1] - ifg[0]).days


def intlist_to_string(ifg_date_list, ext=".int"):
    """Convert date pairs to list of string filenames

    Parameters
    ----------
    ifg_date_list : iteratble of (datetime, datetime)

    ext : str
        Filename extension to strip (Default value = ".int")

    Returns
    -------

    """
    return [
        "{}_{}{ext}".format(a.strftime(DATE_FMT), b.strftime(DATE_FMT), ext=ext)
        for a, b in ifg_date_list
    ]


def anisotropy_var(azi_psd1d):
    return np.var(azi_psd1d)


def anisotropy_A(azi_psd1d):
    return 2 * np.ptp(azi_psd1d)


def rms_noise_vs_dist(image, resolution, crop=True):
    """Get the rms noise vs distance for a given image

    Parameters
    ----------
    image : 2D or 3D ndarray
        Noise image to get rms noise vs distance for
    resolution : float
        Resolution of the image (in meters)
    crop :
         (Default value = True)

    Returns
    -------
    bins : ndarray
        The 1D array of bins.
        `bins` will always be 1D, even if passing multiple images.
    rms_noise_vs_dist : ndarray
        The array of rms noise at each distance bin
        If image is 3D, the rms noise is return for each image.
    """
    from scipy import ndimage

    if image.ndim > 2:
        out = [rms_noise_vs_dist(s, resolution) for s in image]
        bins_arr, rms_noise_arr = zip(*out)
        return bins_arr[0], np.array(rms_noise_arr)

    if crop:
        image = crop_square_data(image)

    h, w = image.shape[-2:]
    hc, wc = h // 2, w // 2
    image_centered = image - image[hc, wc]
    # Only extend out to shortest dimension
    # this will miss power contributions in 'corners' r > min(hc, wc)
    num_r = min(hc, wc)

    # create an array of integer radial distances from the center
    Y, X = np.ogrid[0:h, 0:w]
    # These act as "labels" in the ndimage summation
    r = np.hypot(X - wc, Y - hc).round().astype(int)
    # Ignore nans:
    nan_idxs = np.isnan(image_centered)
    r[nan_idxs] = -9999

    # average all image pixels with label 'r' for 0<=r<=wc
    noise_squared = ndimage.mean(image_centered**2, r, index=np.arange(0, num_r))
    rms_noise = np.sqrt(noise_squared)

    bins = r[hc, wc:] * resolution  # for the actual radial distances in image
    assert len(bins) == len(rms_noise)
    return bins, rms_noise


def take_looks(arr, row_looks, col_looks, separate_complex=False, **kwargs):
    """Downsample a numpy matrix by summing blocks of (row_looks, col_looks)

    Cuts off values if the size isn't divisible by num looks

    Parameters
    ----------
    arr : ndarray
        2D array of an image
    row_looks : int
        the reduction rate in row direction
    col_looks : int
        the reduction rate in col direction
    separate_complex : bool
        take looks on magnitude and phase separately
        Better to preserve the look of the magnitude (Default value = False)


    Returns
    -------
        ndarray, size = ceil(rows / row_looks, cols / col_looks)

    Notes
    -----
    For complex data, looks on the magnitude are done separately
    from looks on the phase

    """
    if row_looks == 1 and col_looks == 1:
        return arr
    if arr.ndim == 3:
        return np.stack(
            [
                take_looks(
                    a, row_looks, col_looks, separate_complex=separate_complex, **kwargs
                )
                for a in arr
            ]
        )
    if np.iscomplexobj(arr) and separate_complex:
        mag_looked = take_looks(np.abs(arr), row_looks, col_looks)
        phase_looked = take_looks(np.angle(arr), row_looks, col_looks)
        return mag_looked * np.exp(1j * phase_looked)

    rows, cols = arr.shape
    new_rows = rows // row_looks
    new_cols = cols // col_looks

    row_cutoff = rows % row_looks
    col_cutoff = cols % col_looks

    if row_cutoff != 0:
        arr = arr[:-row_cutoff, :]
    if col_cutoff != 0:
        arr = arr[:, :-col_cutoff]
    # For taking the mean, treat integers as floats
    if np.issubdtype(arr.dtype, np.integer):
        arr = arr.astype("float")

    return np.mean(
        np.reshape(arr, (new_rows, row_looks, new_cols, col_looks)), axis=(3, 1)
    )


def block_iterator(arr_shape, block_shape, overlaps=(0, 0), start_offsets=(0, 0)):
    """Iterator to get indexes for accessing blocks of a raster

    Parameters
    ----------
    arr_shape :
        (num_rows, num_cols), full size of array to access
    block_shape :
        (height, width), size of accessing blocks
    overlaps : tuple[int, int]
        (row_overlap, col_overlap), number of pixels to re-include
        after sliding the block (default (0, 0))
        
    start_offsets :
         (Default value = (0)

    Yields
    -------
    Iterator of ((row_start, row_stop), (col_start, col_stop))

    Examples
    --------
    >>> list(block_iterator((180, 250), (100, 100)))
    [((0, 100), (0, 100)), ((0, 100), (100, 200)), ((0, 100), (200, 250)), \
((100, 180), (0, 100)), ((100, 180), (100, 200)), ((100, 180), (200, 250))]
    >>> list(block_iterator((180, 250), (100, 100), overlaps=(10, 10)))
    [((0, 100), (0, 100)), ((0, 100), (90, 190)), ((0, 100), (180, 250)), \
((90, 180), (0, 100)), ((90, 180), (90, 190)), ((90, 180), (180, 250))]

    """
    rows, cols = arr_shape
    row_off, col_off = start_offsets
    row_overlap, col_overlap = overlaps
    height, width = block_shape

    if height is None:
        height = rows
    if width is None:
        width = cols

    # Check we're not moving backwards with the overlap:
    if row_overlap >= height:
        raise ValueError(f"row_overlap {row_overlap} must be less than {height}")
    if col_overlap >= width:
        raise ValueError(f"col_overlap {col_overlap} must be less than {width}")
    while row_off < rows:
        while col_off < cols:
            row_end = min(row_off + height, rows)  # Dont yield something OOB
            col_end = min(col_off + width, cols)
            yield ((row_off, row_end), (col_off, col_end))

            col_off += width
            if col_off < cols:  # dont bring back if already at edge
                col_off -= col_overlap

        row_off += height
        if row_off < rows:
            row_off -= row_overlap
        col_off = 0



import os
import shutil

import numpy as np

from troposim import deramp, turbulence


def create_deramped_zarr(
    deramped_file,
    avg_slc_file,
    mask=None,
    ds="igrams",
    units="rad",
    scaling=1.0,
    overwrite=False,
):
    """Create a deramped zarr file from a list of average SLCs

    Parameters
    ----------
    deramped_file : str
        output file name to save deramped data to
    avg_slc_file : str
        path to the average SLC file created by trodi package
    mask : ndarray
        (Default value = None)
    ds : str
        (Default value = "igrams")
    units : str
        (Default value = "rad")
    scaling : float
        (Default value = 1.0)
    overwrite : bool
        (Default value = False)
    """
    try:
        import xarray as xr
        from numcodecs import Blosc
    except ImportError:
        print("xarray and numcodecs must be installed to use this function")
        raise

    if os.path.exists(deramped_file):
        print(f"{deramped_file} already exists")
        if not overwrite:
            return deramped_file
        else:
            print(f"Removing {deramped_file}")
            shutil.rmtree(deramped_file)

    avg_slcs = xr.open_dataset(avg_slc_file)
    outstack = deramp.remove_ramp(
        avg_slcs[ds].data, copy=True, deramp_order=2, mask=mask
    )
    if mask is not None:
        if mask.ndim > 2:
            outstack[mask] = np.nan
        else:
            outstack[:, mask] = np.nan

    avg_slcs[ds].data = outstack * scaling

    compressor = Blosc(cname="zstd", clevel=3, shuffle=Blosc.BITSHUFFLE)
    avg_slcs[ds].attrs["units"] = units
    print(f"saving to {deramped_file}")
    avg_slcs.to_zarr(deramped_file, encoding={ds: {"compressor": compressor}})


# Get all 1d PSD curves from the average ifgs in a region
def get_all_1d_psd_curves(
    avg_ifgs,
    resolution=180,
    min_date=None,
    max_date=None,
    freq0=1e-4,
    deg=3,
    save_dir=".",
    load=True,
):
    """Get all 1D radially averaged power spectrum curves averaged for a region

    Parameters
    ----------
    avg_ifgs : xr.DataArray
        average interferograms
    resolution : float
        spatial resolution of input data in meters (Default value = 180)
    min_date : datetime
        minimum date to include in the analysis (Default value = None)
    max_date : datetime
        maximum date to include in the analysis (Default value = None)
    save_dir : str
        directory to save the output to in .npz file (Default value = ".")
    freq0 :
        (Default value = 1e-4)
    deg :
        (Default value = 3)
    load :
        (Default value = True)
    """
    stack = avg_ifgs.sel(date=slice(min_date, max_date))
    # stack_data = stack.as_numpy()
    # stack_data.load()
    if load:
        stack.load()
    p0_arr, beta_arr, freq_arr, psd1d_arr = turbulence.get_psd(
        image=stack.data,
        resolution=resolution,
        freq0=freq0,
        deg=deg,
        crop=True,
    )
    beta_arr_coeffs = np.array([b.coef for b in beta_arr])
    beta_arr_mean = np.polynomial.Polynomial(np.mean(beta_arr_coeffs, axis=0))

    if save_dir:
        fname = f"psd1d_maxdate_{max_date}.npz" if max_date else "psd1d.npz"
        save_name = os.path.join(save_dir, fname)
        print(f"Saving to {save_name}")
        np.savez(
            save_name,
            p0_arr=p0_arr,
            beta_arr=beta_arr,
            beta_mean=beta_arr_mean,
            freq_arr=freq_arr,
            psd1d_arr=psd1d_arr,
        )
    return p0_arr, beta_arr, freq_arr, psd1d_arr



def get_psd_blocks(
    data, block_size=None, block_step=None, resolution=60.0, freq0=1e-4, deg=3
):
    """For one image, get radially averaged PSD from multiple blocks within image
    Crops into subsets, calls `get_psd` on each

    Parameters
    ----------
    data : 2D ndarray
        displacement in m.
    data : 2D ndarray
        displacement in m.
        resolution (float), spatial resolution of input data in meters
    block_size : float
        size of block side, in meters (Default value = None)
    block_step : float
        amount to shift the block window, in m (Default value = None)
    block_step : float
        amount to shift the block window, in m
        freq0 (float), reference spatial freqency in cycle / m. (Default value = None)
    resolution :
        (Default value = 60.0)
    freq0 :
        (Default value = 1e-4)
    deg :
        (Default value = 3)

    Returns
    -------


    """
    nrow, ncol = data.shape
    if block_size is None:
        block_pix = min(nrow, ncol) // 2
    else:
        block_pix = int(block_size / resolution)
    if block_step is None:
        step = min(nrow, ncol) // 4
    else:
        step = int(block_step / resolution)

    p0_hat_arr, beta_hat_list, psd1d_arr = [], [], []
    freq = None
    row, col = 0, 0
    while row + block_pix - 1 < data.shape[0]:
        while col + block_pix - 1 < data.shape[1]:
            # print(row, col, block_size, block_pix, block_step, step)
            block = data[row : row + block_pix, col : col + block_pix]
            if block.shape != (block_pix, block_pix):
                continue
            p0_hat, beta_hat, freq, psd1d = turbulence.get_psd(
                block,
                resolution=resolution,
                freq0=freq0,
                crop=False,
                deg=deg,
            )
            p0_hat_arr.append(p0_hat)
            beta_hat_list.append(beta_hat)
            psd1d_arr.append(psd1d)
            col += step
        col = 0
        row += step
    return np.array(p0_hat_arr), np.array(beta_hat_list), freq, np.array(psd1d_arr)


def fractal_dimension(beta):
    """Convert the beta slope to D, fractal dimension (D2 is from Hanssen Section 4.7"""
    return (7.0 - beta + 1.0) / 2.0


def debug(surf):
    return (
        f"min:{np.min(surf):.2g}, max:{np.max(surf):.2g}, "
        f"ptp:{np.ptp(surf):.2g}, mean:{np.mean(surf):.2g}"
    )


def average_psd_azimuth(
    psd2d=None, image=None, resolution=60.0, dTheta=30, r_min=0, r_max=np.inf
):
    """Get 1D power spectrum averaged by angular bin (azimuthal average)

    Parameters
    ----------
    psd2d : ndarray
        2D power spectral density, size (N, N)
        non-square images sized (r, c) will use N = min(r, c) (Default value = None)
    image : ndarray, optional
        if psd2d=None, pass original image to transform
        (Default value = None)
    dTheta : float
        spacing between angle bins, in degrees. default=30.
    r_min : float
        optional, to limit the radius where the averaging takes place,
        pass a minimum radius (in pixels) (Default value = 0)
    r_max : float
        optional, maximum radius (in pixels) where the averaging takes place
        (Default value = np.inf)
    resolution :
        (Default value = 60.0)

    Returns
    -------
    angles : ndarray
        angles in degrees from 0 to 180
    psd1d : ndarray
        1D power spectral density, size same as angles
    """
    from scipy import ndimage
    if psd2d is None:
        if image is None:
            raise ValueError("need either psd, or pre-transformed image")
        psd2d = turbulence.get_psd2d(image, resolution=resolution)
    theta = _get_theta_sectors(psd2d, dTheta, r_min=r_min, r_max=r_max)

    # use all psd2d pixels with label 'theta' (0 to 180) between (r_min, r_max)
    # psd is symmetric for real data
    angles = np.arange(0, 180, int(round(dTheta)))
    psd1d = ndimage.sum(psd2d, theta, index=angles)

    # normalize each sector to the total sector power
    psd1d /= np.sum(psd1d)

    return angles, psd1d


def _get_theta_sectors(psd2d, dTheta, r_min=0, r_max=np.inf):
    """Helper to make a mask of azimuthal angles of image"""
    h, w = psd2d.shape
    hc, wc = h // 2, w // 2

    # note that displaying PSD as image inverts Y axis
    # create an array of integer angular slices of dTheta
    Y, X = np.ogrid[0:h, 0:w]
    theta = np.rad2deg(np.arctan2(-(Y - hc), (X - wc)))
    theta = np.mod(theta + dTheta / 2 + 360, 360)
    theta = dTheta * (theta // dTheta)
    theta = theta.round().astype(np.int)

    # mask below r_min and above r_max by setting to -999
    R = np.hypot((Y - hc), (X - wc))
    mask = np.logical_and(R >= r_min, R < r_max)
    theta = np.where(mask, theta, np.nan)
    return theta
